from ._pattern_detector import PatternDetector
from ._pattern_detector_controller import PatternDetectorController
from ._pattern_detector_writer import PatternDetectorWriter
from ._pattern_generator import DATA_PATH, SUM_PATH, PatternGenerator

__all__ = [
    "PatternDetector",
    "PatternDetectorController",
    "PatternDetectorWriter",
    "DATA_PATH",
    "SUM_PATH",
    "PatternGenerator",
]



import asyncio

from ophyd_async.core import DetectorController, PathProvider, TriggerInfo

from ._pattern_generator import PatternGenerator


class PatternDetectorController(DetectorController):
    def __init__(
        self,
        pattern_generator: PatternGenerator,
        path_provider: PathProvider,
        exposure: float = 0.1,
    ) -> None:
        self.pattern_generator: PatternGenerator = pattern_generator
        self.pattern_generator.set_exposure(exposure)
        self.path_provider: PathProvider = path_provider
        self.task: asyncio.Task | None = None
        super().__init__()

    async def prepare(self, trigger_info: TriggerInfo):
        self._trigger_info = trigger_info
        if self._trigger_info.livetime is None:
            self._trigger_info.livetime = 0.01
        self.period: float = self._trigger_info.livetime + self.get_deadtime(
            trigger_info.livetime
        )

    async def arm(self):
        assert self._trigger_info.livetime
        assert self.period
        self.task = asyncio.create_task(
            self._coroutine_for_image_writing(
                self._trigger_info.livetime,
                self.period,
                self._trigger_info.total_number_of_triggers,
            )
        )

    async def wait_for_idle(self):
        if self.task:
            await self.task

    async def disarm(self):
        if self.task and not self.task.done():
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError:
                pass
        self.task = None

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.001

    async def _coroutine_for_image_writing(
        self, exposure: float, period: float, frames_number: int
    ):
        for _ in range(frames_number):
            self.pattern_generator.set_exposure(exposure)
            await asyncio.sleep(period)
            await self.pattern_generator.write_image_to_file()



import warnings
from collections.abc import Awaitable, Callable, Generator, Sequence
from contextlib import contextmanager
from enum import Enum
from typing import Any, cast

from bluesky.protocols import HasHints, Hints, Reading
from event_model import DataKey

from ._device import Device, DeviceVector
from ._protocol import AsyncConfigurable, AsyncReadable, AsyncStageable
from ._signal import SignalR
from ._status import AsyncStatus
from ._utils import merge_gathered_dicts


class StandardReadableFormat(Enum):
    """Declare how a `Device` should contribute to the `StandardReadable` verbs."""

    #: Detect which verbs the child supports and contribute to:
    #:
    #: - ``read()``, ``describe()`` if it is `bluesky.protocols.Readable`
    #: - ``read_configuration()``, ``describe_configuration()`` if it is
    #:   `bluesky.protocols.Configurable`
    #: - ``stage()``, ``unstage()`` if it is `bluesky.protocols.Stageable`
    #: - ``hints`` if it `bluesky.protocols.HasHints`
    CHILD = "CHILD"
    #: Contribute the `Signal` value to ``read_configuration()`` and
    #: ``describe_configuration()``
    CONFIG_SIGNAL = "CONFIG_SIGNAL"
    #: Contribute the monitored `Signal` value to ``read()`` and ``describe()``` and
    #: put the signal name in ``hints``
    HINTED_SIGNAL = "HINTED_SIGNAL"
    #: Contribute the uncached `Signal` value to ``read()`` and ``describe()```
    UNCACHED_SIGNAL = "UNCACHED_SIGNAL"
    #: Contribute the uncached `Signal` value to ``read()`` and ``describe()``` and
    #: put the signal name in ``hints``
    HINTED_UNCACHED_SIGNAL = "HINTED_UNCACHED_SIGNAL"

    def __call__(self, parent: Device, child: Device):
        if not isinstance(parent, StandardReadable):
            raise TypeError(f"Expected parent to be StandardReadable, got {parent}")
        parent.add_readables([child], self)


# Back compat
class _WarningMatcher:
    def __init__(self, name: str, target: StandardReadableFormat):
        self._name = name
        self._target = target

    def __eq__(self, value: object) -> bool:
        warnings.warn(
            DeprecationWarning(
                f"Use `StandardReadableFormat.{self._target.name}` "
                f"instead of `{self._name}`"
            ),
            stacklevel=2,
        )
        return value == self._target


def _compat_format(name: str, target: StandardReadableFormat) -> StandardReadableFormat:
    return cast(StandardReadableFormat, _WarningMatcher(name, target))


ConfigSignal = _compat_format("ConfigSignal", StandardReadableFormat.CONFIG_SIGNAL)
HintedSignal: Any = _compat_format("HintedSignal", StandardReadableFormat.HINTED_SIGNAL)
HintedSignal.uncached = _compat_format(
    "HintedSignal.uncached", StandardReadableFormat.HINTED_UNCACHED_SIGNAL
)


class StandardReadable(
    Device, AsyncReadable, AsyncConfigurable, AsyncStageable, HasHints
):
    """Device that owns its children and provides useful default behavior.

    - When its name is set it renames child Devices
    - Signals can be registered for read() and read_configuration()
    - These signals will be subscribed for read() between stage() and unstage()
    """

    # These must be immutable types to avoid accidental sharing between
    # different instances of the class
    _describe_config_funcs: tuple[Callable[[], Awaitable[dict[str, DataKey]]], ...] = ()
    _read_config_funcs: tuple[Callable[[], Awaitable[dict[str, Reading]]], ...] = ()
    _describe_funcs: tuple[Callable[[], Awaitable[dict[str, DataKey]]], ...] = ()
    _read_funcs: tuple[Callable[[], Awaitable[dict[str, Reading]]], ...] = ()
    _stageables: tuple[AsyncStageable, ...] = ()
    _has_hints: tuple[HasHints, ...] = ()

    @AsyncStatus.wrap
    async def stage(self) -> None:
        for sig in self._stageables:
            await sig.stage().task

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        for sig in self._stageables:
            await sig.unstage().task

    async def describe_configuration(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts(
            [func() for func in self._describe_config_funcs]
        )

    async def read_configuration(self) -> dict[str, Reading]:
        return await merge_gathered_dicts([func() for func in self._read_config_funcs])

    async def describe(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts([func() for func in self._describe_funcs])

    async def read(self) -> dict[str, Reading]:
        return await merge_gathered_dicts([func() for func in self._read_funcs])

    @property
    def hints(self) -> Hints:
        hints: Hints = {}
        for new_hint in self._has_hints:
            # Merge the existing and new hints, based on the type of the value.
            # This avoids default dict merge behaviour that overrides the values;
            # we want to combine them when they are Sequences, and ensure they are
            # identical when string values.
            for key, value in new_hint.hints.items():
                if isinstance(value, str):
                    if key in hints:
                        assert (
                            hints[key] == value  # type: ignore[literal-required]
                        ), f"Hints key {key} value may not be overridden"
                    else:
                        hints[key] = value  # type: ignore[literal-required]
                elif isinstance(value, Sequence):
                    if key in hints:
                        for new_val in value:
                            assert (
                                new_val not in hints[key]  # type: ignore[literal-required]
                            ), f"Hint {key} {new_val} overrides existing hint"
                        hints[key] = (  # type: ignore[literal-required]
                            hints[key] + value  # type: ignore[literal-required]
                        )
                    else:
                        hints[key] = value  # type: ignore[literal-required]
                else:
                    raise TypeError(
                        f"{new_hint.name}: Unknown type for value '{value}' "
                        f" for key '{key}'"
                    )

        return hints

    @contextmanager
    def add_children_as_readables(
        self,
        format: StandardReadableFormat = StandardReadableFormat.CHILD,
    ) -> Generator[None, None, None]:
        """Context manager that calls `add_readables` on child Devices added within.

        Scans ``self.children()`` on entry and exit to context manager, and calls
        `add_readables` on any that are added with the provided
        `StandardReadableFormat`.
        """

        dict_copy = dict(self.children())

        yield

        # Set symmetric difference operator gives all newly added keys
        new_dict = dict(self.children())
        new_keys = dict_copy.keys() ^ new_dict.keys()
        new_values = [new_dict[key] for key in new_keys]

        flattened_values = []
        for value in new_values:
            if isinstance(value, DeviceVector):
                flattened_values.extend(value.values())
            else:
                flattened_values.append(value)

        new_devices = list(filter(lambda x: isinstance(x, Device), flattened_values))
        self.add_readables(new_devices, format)

    def add_readables(
        self,
        devices: Sequence[Device],
        format: StandardReadableFormat = StandardReadableFormat.CHILD,
    ) -> None:
        """Add devices to contribute to various bluesky verbs.

        Use output from the given devices to contribute to the verbs of the following
        interfaces:

        - `bluesky.protocols.Readable`
        - `bluesky.protocols.Configurable`
        - `bluesky.protocols.Stageable`
        - `bluesky.protocols.HasHints`

        Parameters
        ----------
        devices:
            The devices to be added
        format:
            Determines which of the devices functions are added to which verb as per the
            `StandardReadableFormat` documentation
        """

        for device in devices:
            match format:
                case StandardReadableFormat.CHILD:
                    if isinstance(device, AsyncConfigurable):
                        self._describe_config_funcs += (device.describe_configuration,)
                        self._read_config_funcs += (device.read_configuration,)
                    if isinstance(device, AsyncReadable):
                        self._describe_funcs += (device.describe,)
                        self._read_funcs += (device.read,)
                    if isinstance(device, AsyncStageable):
                        self._stageables += (device,)
                    if isinstance(device, HasHints):
                        self._has_hints += (device,)
                case StandardReadableFormat.CONFIG_SIGNAL:
                    assert isinstance(device, SignalR), f"{device} is not a SignalR"
                    self._describe_config_funcs += (device.describe,)
                    self._read_config_funcs += (device.read,)
                case StandardReadableFormat.HINTED_SIGNAL:
                    assert isinstance(device, SignalR), f"{device} is not a SignalR"
                    self._describe_funcs += (device.describe,)
                    self._read_funcs += (device.read,)
                    self._stageables += (device,)
                    self._has_hints += (_HintsFromName(device),)
                case StandardReadableFormat.UNCACHED_SIGNAL:
                    assert isinstance(device, SignalR), f"{device} is not a SignalR"
                    self._describe_funcs += (device.describe,)
                    self._read_funcs += (_UncachedRead(device),)
                case StandardReadableFormat.HINTED_UNCACHED_SIGNAL:
                    assert isinstance(device, SignalR), f"{device} is not a SignalR"
                    self._describe_funcs += (device.describe,)
                    self._read_funcs += (_UncachedRead(device),)
                    self._has_hints += (_HintsFromName(device),)


class _UncachedRead:
    def __init__(self, signal: SignalR) -> None:
        self.signal = signal

    async def __call__(self) -> dict[str, Reading]:
        return await self.signal.read(cached=False)


class _HintsFromName(HasHints):
    def __init__(self, device: Device) -> None:
        self.device = device

    @property
    def name(self) -> str:
        return self.device.name

    @property
    def hints(self) -> Hints:
        fields = [self.name] if self.name else []
        return {"fields": fields}



from ._core_io import (
    ADBaseIO,
    DetectorState,
    NDArrayBaseIO,
    NDFileHDFIO,
    NDPluginStatsIO,
)
from ._core_logic import (
    DEFAULT_GOOD_STATES,
    ADBaseDatasetDescriber,
    set_exposure_time_and_acquire_period_if_supplied,
    start_acquiring_driver_and_ensure_status,
)
from ._hdf_writer import ADHDFWriter
from ._single_trigger import SingleTriggerDetector
from ._utils import (
    ADBaseDataType,
    FileWriteMode,
    ImageMode,
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    NDAttributePvDbrType,
    stop_busy_record,
)

__all__ = [
    "ADBaseIO",
    "DetectorState",
    "NDArrayBaseIO",
    "NDFileHDFIO",
    "NDPluginStatsIO",
    "DEFAULT_GOOD_STATES",
    "ADBaseDatasetDescriber",
    "set_exposure_time_and_acquire_period_if_supplied",
    "start_acquiring_driver_and_ensure_status",
    "ADHDFWriter",
    "SingleTriggerDetector",
    "ADBaseDataType",
    "FileWriteMode",
    "ImageMode",
    "NDAttributePv",
    "NDAttributeParam",
    "NDAttributeDataType",
    "stop_busy_record",
    "NDAttributePvDbrType",
]



from __future__ import annotations

from typing import TypeVar
from unittest.mock import Mock

from ophyd_async.core import Device, DeviceConnector, DeviceFiller
from ophyd_async.tango.signal import (
    TangoSignalBackend,
    infer_python_type,
    infer_signal_type,
)
from tango import DeviceProxy as DeviceProxy
from tango.asyncio import DeviceProxy as AsyncDeviceProxy

T = TypeVar("T")


class TangoDevice(Device):
    """
    General class for TangoDevices. Extends Device to provide attributes for Tango
    devices.

    Parameters
    ----------
    trl: str
        Tango resource locator, typically of the device server.
    device_proxy: Optional[Union[AsyncDeviceProxy, SyncDeviceProxy]]
        Asynchronous or synchronous DeviceProxy object for the device. If not provided,
        an asynchronous DeviceProxy object will be created using the trl and awaited
        when the device is connected.
    """

    trl: str = ""
    proxy: DeviceProxy | None = None
    _polling: tuple[bool, float, float | None, float | None] = (False, 0.1, None, 0.1)
    _signal_polling: dict[str, tuple[bool, float, float, float]] = {}
    _poll_only_annotated_signals: bool = True

    def __init__(
        self,
        trl: str | None = None,
        device_proxy: DeviceProxy | None = None,
        name: str = "",
    ) -> None:
        connector = TangoDeviceConnector(
            trl=trl,
            device_proxy=device_proxy,
            polling=self._polling,
            signal_polling=self._signal_polling,
        )
        super().__init__(name=name, connector=connector)


def tango_polling(
    polling: tuple[float, float, float]
    | dict[str, tuple[float, float, float]]
    | None = None,
    signal_polling: dict[str, tuple[float, float, float]] | None = None,
):
    """
    Class decorator to configure polling for Tango devices.

    This decorator allows for the configuration of both device-level and signal-level
    polling for Tango devices. Polling is useful for device servers that do not support
    event-driven updates.

    Parameters
    ----------
    polling : Optional[Union[Tuple[float, float, float],
              Dict[str, Tuple[float, float, float]]]], optional
        Device-level polling configuration as a tuple of three floats representing the
        polling interval, polling timeout, and polling delay. Alternatively,
        a dictionary can be provided to specify signal-level polling configurations
        directly.
    signal_polling : Optional[Dict[str, Tuple[float, float, float]]], optional
        Signal-level polling configuration as a dictionary where keys are signal names
        and values are tuples of three floats representing the polling interval, polling
        timeout, and polling delay.
    """
    if isinstance(polling, dict):
        signal_polling = polling
        polling = None

    def decorator(cls):
        if polling is not None:
            cls._polling = (True, *polling)
        if signal_polling is not None:
            cls._signal_polling = {k: (True, *v) for k, v in signal_polling.items()}
        return cls

    return decorator


class TangoDeviceConnector(DeviceConnector):
    def __init__(
        self,
        trl: str | None,
        device_proxy: DeviceProxy | None,
        polling: tuple[bool, float, float | None, float | None],
        signal_polling: dict[str, tuple[bool, float, float, float]],
    ) -> None:
        self.trl = trl
        self.proxy = device_proxy
        self._polling = polling
        self._signal_polling = signal_polling

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            self.filler = DeviceFiller(
                device=device,
                signal_backend_factory=TangoSignalBackend,
                device_connector_factory=lambda: TangoDeviceConnector(
                    None, None, (False, 0.1, None, None), {}
                ),
            )
            list(self.filler.create_devices_from_annotations(filled=False))
            list(self.filler.create_signals_from_annotations(filled=False))
            self.filler.check_created()

    async def connect(
        self, device: Device, mock: bool | Mock, timeout: float, force_reconnect: bool
    ) -> None:
        if mock:
            # Make 2 entries for each DeviceVector
            self.filler.create_device_vector_entries_to_mock(2)
        else:
            if self.trl and self.proxy is None:
                self.proxy = await AsyncDeviceProxy(self.trl)
            elif self.proxy and not self.trl:
                self.trl = self.proxy.name()
            else:
                raise TypeError("Neither proxy nor trl supplied")

            children = sorted(
                set()
                .union(self.proxy.get_attribute_list())
                .union(self.proxy.get_command_list())
            )
            for name in children:
                # TODO: strip attribute name
                full_trl = f"{self.trl}/{name}"
                signal_type = await infer_signal_type(full_trl, self.proxy)
                if signal_type:
                    backend = self.filler.fill_child_signal(name, signal_type)
                    backend.datatype = await infer_python_type(full_trl, self.proxy)
                    backend.set_trl(full_trl)
                    if polling := self._signal_polling.get(name, ()):
                        backend.set_polling(*polling)
                        backend.allow_events(False)
                    elif self._polling[0]:
                        backend.set_polling(*self._polling)
                        backend.allow_events(False)
            # Check that all the requested children have been filled
            self.filler.check_filled(f"{self.trl}: {children}")
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect(device, mock, timeout, force_reconnect)



"""Module which defines abstract classes to work with detectors"""

import asyncio
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Iterator, Sequence
from functools import cached_property

from bluesky.protocols import (
    Collectable,
    Flyable,
    Preparable,
    Reading,
    Stageable,
    StreamAsset,
    Triggerable,
    WritesStreamAssets,
)
from event_model import DataKey
from pydantic import BaseModel, Field, NonNegativeInt, computed_field

from ._device import Device, DeviceConnector
from ._protocol import AsyncConfigurable, AsyncReadable
from ._signal import SignalR
from ._status import AsyncStatus, WatchableAsyncStatus
from ._utils import DEFAULT_TIMEOUT, StrictEnum, WatcherUpdate, merge_gathered_dicts


class DetectorTrigger(StrictEnum):
    """Type of mechanism for triggering a detector to take frames"""

    #: Detector generates internal trigger for given rate
    internal = "internal"
    #: Expect a series of arbitrary length trigger signals
    edge_trigger = "edge_trigger"
    #: Expect a series of constant width external gate signals
    constant_gate = "constant_gate"
    #: Expect a series of variable width external gate signals
    variable_gate = "variable_gate"


class TriggerInfo(BaseModel):
    """Minimal set of information required to setup triggering on a detector"""

    #: Number of triggers that will be sent, (0 means infinite) Can be:
    #  - A single integer or
    #  - A list of integers for multiple triggers
    # Example for tomography: TriggerInfo(number=[2,3,100,3])
    #:     This would trigger:
    #:     - 2 times for dark field images
    #:     - 3 times for initial flat field images
    #:     - 100 times for projections
    #:     - 3 times for final flat field images
    number_of_triggers: NonNegativeInt | list[NonNegativeInt]
    #: Sort of triggers that will be sent
    trigger: DetectorTrigger = Field(default=DetectorTrigger.internal)
    #: What is the minimum deadtime between triggers
    deadtime: float | None = Field(default=None, ge=0)
    #: What is the maximum high time of the triggers
    livetime: float | None = Field(default=None, ge=0)
    #: What is the maximum timeout on waiting for a frame
    frame_timeout: float | None = Field(default=None, gt=0)
    #: How many triggers make up a single StreamDatum index, to allow multiple frames
    #: from a faster detector to be zipped with a single frame from a slow detector
    #: e.g. if num=10 and multiplier=5 then the detector will take 10 frames,
    #: but publish 2 indices, and describe() will show a shape of (5, h, w)
    multiplier: int = 1

    @computed_field
    @cached_property
    def total_number_of_triggers(self) -> int:
        return (
            sum(self.number_of_triggers)
            if isinstance(self.number_of_triggers, list)
            else self.number_of_triggers
        )


class DetectorController(ABC):
    """
    Classes implementing this interface should hold the logic for
    arming and disarming a detector
    """

    @abstractmethod
    def get_deadtime(self, exposure: float | None) -> float:
        """For a given exposure, how long should the time between exposures be"""

    @abstractmethod
    async def prepare(self, trigger_info: TriggerInfo):
        """
        Do all necessary steps to prepare the detector for triggers.

        Args:
            trigger_info: This is a Pydantic model which contains
                number Expected number of frames.
                trigger Type of trigger for which to prepare the detector. Defaults
                to DetectorTrigger.internal.
                livetime Livetime / Exposure time with which to set up the detector.
                Defaults to None
                if not applicable or the detector is expected to use its previously-set
                exposure time.
                deadtime Defaults to None. This is the minimum deadtime between
                triggers.
                multiplier The number of triggers grouped into a single StreamDatum
                index.
        """

    @abstractmethod
    async def arm(self) -> None:
        """
        Arm the detector
        """

    @abstractmethod
    async def wait_for_idle(self):
        """
        This will wait on the internal _arm_status and wait for it to get disarmed/idle
        """

    @abstractmethod
    async def disarm(self):
        """Disarm the detector, return detector to an idle state"""


class DetectorWriter(ABC):
    """Logic for making a detector write data to somewhere persistent
    (e.g. an HDF5 file)"""

    @abstractmethod
    async def open(self, multiplier: int = 1) -> dict[str, DataKey]:
        """Open writer and wait for it to be ready for data.

        Args:
            multiplier: Each StreamDatum index corresponds to this many
                written exposures

        Returns:
            Output for ``describe()``
        """

    @abstractmethod
    def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        """Yield the index of each frame (or equivalent data point) as it is written"""

    @abstractmethod
    async def get_indices_written(self) -> int:
        """Get the number of indices written"""

    @abstractmethod
    def collect_stream_docs(self, indices_written: int) -> AsyncIterator[StreamAsset]:
        """Create Stream docs up to given number written"""

    @abstractmethod
    async def close(self) -> None:
        """Close writer, blocks until I/O is complete"""


class StandardDetector(
    Device,
    Stageable,
    AsyncConfigurable,
    AsyncReadable,
    Triggerable,
    Preparable,
    Flyable,
    Collectable,
    WritesStreamAssets,
):
    """
    Useful detector base class for step and fly scanning detectors.
    Aggregates controller and writer logic together.
    """

    def __init__(
        self,
        controller: DetectorController,
        writer: DetectorWriter,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
        connector: DeviceConnector | None = None,
    ) -> None:
        """
        Constructor

        Args:
            controller: Logic for arming and disarming the detector
            writer: Logic for making the detector write persistent data
            config_sigs: Signals to read when describe and read
            configuration are called. Defaults to ().
            name: Device name. Defaults to "".
        """
        self._controller = controller
        self._writer = writer
        self._describe: dict[str, DataKey] = {}
        self._config_sigs = list(config_sigs)
        # For prepare
        self._arm_status: AsyncStatus | None = None
        self._trigger_info: TriggerInfo | None = None
        # For kickoff
        self._watchers: list[Callable] = []
        self._fly_status: WatchableAsyncStatus | None = None
        self._fly_start: float | None = None
        self._frames_to_complete: int = 0
        # Represents the total number of frames that will have been completed at the
        # end of the next `complete`.
        self._completable_frames: int = 0
        self._number_of_triggers_iter: Iterator[int] | None = None
        self._initial_frame: int = 0
        super().__init__(name, connector=connector)

    @property
    def controller(self) -> DetectorController:
        return self._controller

    @property
    def writer(self) -> DetectorWriter:
        return self._writer

    @AsyncStatus.wrap
    async def stage(self) -> None:
        # Disarm the detector, stop file writing.
        await self._check_config_sigs()
        await asyncio.gather(self.writer.close(), self.controller.disarm())
        self._trigger_info = None

    async def _check_config_sigs(self):
        """Checks configuration signals are named and connected."""
        for signal in self._config_sigs:
            if signal.name == "":
                raise Exception(
                    "config signal must be named before it is passed to the detector"
                )
            try:
                await signal.get_value()
            except NotImplementedError as e:
                raise Exception(
                    f"config signal {signal.name} must be connected before it is "
                    + "passed to the detector"
                ) from e

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        # Stop data writing.
        await asyncio.gather(self.writer.close(), self.controller.disarm())

    async def read_configuration(self) -> dict[str, Reading]:
        return await merge_gathered_dicts(sig.read() for sig in self._config_sigs)

    async def describe_configuration(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts(sig.describe() for sig in self._config_sigs)

    async def read(self) -> dict[str, Reading]:
        # All data is in StreamResources, not Events, so nothing to output here
        return {}

    async def describe(self) -> dict[str, DataKey]:
        return self._describe

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        if self._trigger_info is None:
            await self.prepare(
                TriggerInfo(
                    number_of_triggers=1,
                    trigger=DetectorTrigger.internal,
                    deadtime=None,
                    livetime=None,
                    frame_timeout=None,
                )
            )
        assert self._trigger_info
        assert self._trigger_info.trigger is DetectorTrigger.internal
        # Arm the detector and wait for it to finish.
        indices_written = await self.writer.get_indices_written()
        await self.controller.arm()
        await self.controller.wait_for_idle()
        end_observation = indices_written + 1

        async for index in self.writer.observe_indices_written(
            DEFAULT_TIMEOUT
            + (self._trigger_info.livetime or 0)
            + (self._trigger_info.deadtime or 0)
        ):
            if index >= end_observation:
                break

    @AsyncStatus.wrap
    async def prepare(self, value: TriggerInfo) -> None:
        """
        Arm detector.

        Prepare the detector with trigger information. This is determined at and passed
        in from the plan level.

        This currently only prepares detectors for flyscans and stepscans just use the
        trigger information determined in trigger.

        To do: Unify prepare to be use for both fly and step scans.

        Args:
            value: TriggerInfo describing how to trigger the detector
        """
        if value.trigger != DetectorTrigger.internal:
            assert (
                value.deadtime
            ), "Deadtime must be supplied when in externally triggered mode"
        if value.deadtime:
            required = self.controller.get_deadtime(value.livetime)
            assert required <= value.deadtime, (
                f"Detector {self.controller} needs at least {required}s deadtime, "
                f"but trigger logic provides only {value.deadtime}s"
            )
        self._trigger_info = value
        self._number_of_triggers_iter = iter(
            self._trigger_info.number_of_triggers
            if isinstance(self._trigger_info.number_of_triggers, list)
            else [self._trigger_info.number_of_triggers]
        )
        self._initial_frame = await self.writer.get_indices_written()
        self._describe, _ = await asyncio.gather(
            self.writer.open(value.multiplier), self.controller.prepare(value)
        )
        if value.trigger != DetectorTrigger.internal:
            await self.controller.arm()
            self._fly_start = time.monotonic()

    @AsyncStatus.wrap
    async def kickoff(self):
        if self._trigger_info is None or self._number_of_triggers_iter is None:
            raise RuntimeError("Prepare must be called before kickoff!")
        try:
            self._frames_to_complete = next(self._number_of_triggers_iter)
            self._completable_frames += self._frames_to_complete
        except StopIteration as err:
            raise RuntimeError(
                f"Kickoff called more than the configured number of "
                f"{self._trigger_info.total_number_of_triggers} iteration(s)!"
            ) from err

    @WatchableAsyncStatus.wrap
    async def complete(self):
        assert self._trigger_info
        indices_written = self.writer.observe_indices_written(
            self._trigger_info.frame_timeout
            or (
                DEFAULT_TIMEOUT
                + (self._trigger_info.livetime or 0)
                + (self._trigger_info.deadtime or 0)
            )
        )
        try:
            async for index in indices_written:
                yield WatcherUpdate(
                    name=self.name,
                    current=index,
                    initial=self._initial_frame,
                    target=self._frames_to_complete,
                    unit="",
                    precision=0,
                    time_elapsed=time.monotonic() - self._fly_start
                    if self._fly_start
                    else None,
                )
                if index >= self._frames_to_complete:
                    break
        finally:
            await indices_written.aclose()
            if self._completable_frames >= self._trigger_info.total_number_of_triggers:
                self._completable_frames = 0
                self._frames_to_complete = 0
                self._number_of_triggers_iter = None
                await self.controller.wait_for_idle()

    async def describe_collect(self) -> dict[str, DataKey]:
        return self._describe

    async def collect_asset_docs(
        self, index: int | None = None
    ) -> AsyncIterator[StreamAsset]:
        # Collect stream datum documents for all indices written.
        # The index is optional, and provided for fly scans, however this needs to be
        # retrieved for step scans.
        if index is None:
            index = await self.writer.get_indices_written()
        async for doc in self.writer.collect_stream_docs(index):
            yield doc

    async def get_index(self) -> int:
        return await self.writer.get_indices_written()



from collections.abc import Sequence
from pathlib import Path

from ophyd_async.core import (
    FilenameProvider,
    PathProvider,
    SignalR,
    StandardDetector,
    StaticFilenameProvider,
    StaticPathProvider,
)

from ._pattern_detector_controller import PatternDetectorController
from ._pattern_detector_writer import PatternDetectorWriter
from ._pattern_generator import PatternGenerator


class PatternDetector(StandardDetector):
    def __init__(
        self,
        path: Path,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        fp: FilenameProvider = StaticFilenameProvider(name)
        self.path_provider: PathProvider = StaticPathProvider(fp, path)
        self.pattern_generator = PatternGenerator()
        writer = PatternDetectorWriter(
            pattern_generator=self.pattern_generator,
            path_provider=self.path_provider,
            name_provider=lambda: self.name,
        )
        controller = PatternDetectorController(
            pattern_generator=self.pattern_generator,
            path_provider=self.path_provider,
        )
        super().__init__(
            controller=controller,
            writer=writer,
            config_sigs=config_sigs,
            name=name,
        )



import inflection
import pytest

from ophyd_async.core import Device, Signal

# Need to import them all so the subclass walking gets all subclasses
# If we forget then the full test suite will find the subclasses, but
# running just this test will only get the ones at the top of this file
from ophyd_async.epics import (
    adaravis,  # noqa
    adcore,
    adkinetix,  # noqa
    adpilatus,  # noqa
    adsimdetector,  # noqa
    advimba,  # noqa
)


def get_rec_subclasses(cls: type):
    yield cls
    for subcls in cls.__subclasses__():
        yield from get_rec_subclasses(subcls)


@pytest.mark.parametrize("cls", list(get_rec_subclasses(adcore.NDArrayBaseIO)))
async def test_regularly_named_attributes(cls: Device):
    io = cls("")
    for name, signal in io.children():
        assert isinstance(signal, Signal)
        # Strip off the ca:// prefix and an _RBV suffix
        pv = signal.source.split("://")[-1].split("_RBV")[0]
        assert inflection.underscore(pv) == name



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import os
import sys
from pathlib import Path
from subprocess import check_output

import requests

import ophyd_async

# -- General configuration ------------------------------------------------
sys.path.insert(0, os.path.abspath("../../src"))
# General information about the project.
project = "ophyd-async"
copyright = "2014, Brookhaven National Lab"

# The full version, including alpha/beta/rc tags.
release = ophyd_async.__version__

# The short X.Y version.
if "+" in release:
    # Not on a tag, use branch name
    root = Path(__file__).absolute().parent.parent
    git_branch = check_output("git branch --show-current".split(), cwd=root)
    version = git_branch.decode().strip()
else:
    version = release

extensions = [
    # for diagrams
    "sphinxcontrib.mermaid",
    # Used for BaseModel autodoc
    "sphinxcontrib.autodoc_pydantic",
    # Use this for generating API docs
    "sphinx.ext.autodoc",
    # Not sure if this is still used?
    "sphinx.ext.doctest",
    # and making summary tables at the top of API docs
    "sphinx.ext.autosummary",
    # This can parse google style docstrings
    "sphinx.ext.napoleon",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
    # Adds the inheritance-diagram generation directive
    "sphinx.ext.inheritance_diagram",
    # Add a copy button to each code block
    "sphinx_copybutton",
    # For the card element
    "sphinx_design",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
    "myst_parser",
    "numpydoc",
]

# So we can use the ::: syntax
myst_enable_extensions = ["colon_fence"]

napoleon_google_docstring = False
napoleon_numpy_docstring = True

# If true, Sphinx will warn about all references where the target cannot
# be found.
# nitpicky = True

# A list of (type, target) tuples (by default empty) that should be ignored when
# generating warnings in "nitpicky mode". Note that type should include the
# domain name if present. Example entries would be ('py:func', 'int') or
# ('envvar', 'LD_LIBRARY_PATH').
nitpick_ignore = [
    # builtins
    ("py:class", "NoneType"),
    ("py:class", "'str'"),
    ("py:class", "'float'"),
    ("py:class", "'int'"),
    ("py:class", "'bool'"),
    ("py:class", "'object'"),
    ("py:class", "'id'"),
    # typing
    ("py:class", "typing_extensions.Literal"),
]

# Order the members by the order they appear in the source code
autodoc_member_order = "bysource"

# Don't inherit docstrings from baseclasses
autodoc_inherit_docstrings = False

# Add some more modules to the top level autosummary
ophyd_async.__all__ += ["sim", "epics", "tango", "fastcs", "plan_stubs"]

# Document only what is in __all__
autosummary_ignore_module_all = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Output graphviz directive produced images in a scalable format
graphviz_output_format = "svg"

# The name of a reST role (builtin or Sphinx extension) to use as the default
# role, that is, for text marked up `like this`
default_role = "any"

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# These patterns also affect html_static_path and html_extra_path
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# Example configuration for intersphinx: refer to the Python standard library.
# This means you can link things like `str` and `asyncio` to the relevant
# docs in the python documentation.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "bluesky": ("https://blueskyproject.io/bluesky/main", None),
    "numpy": ("https://numpy.org/devdocs/", None),
    "databroker": ("https://blueskyproject.io/databroker/", None),
    "event-model": ("https://blueskyproject.io/event-model/main", None),
}

# A dictionary of graphviz graph attributes for inheritance diagrams.
inheritance_graph_attrs = {"rankdir": "TB"}

# Common links that should be available on every page
rst_epilog = """
.. _NSLS: https://www.bnl.gov/nsls2
.. _black: https://github.com/psf/black
.. _ruff: https://beta.ruff.rs/docs/
.. _mypy: http://mypy-lang.org/
.. _pre-commit: https://pre-commit.com/
"""

# Ignore localhost links for periodic check that links in docs are valid
linkcheck_ignore = [r"http://localhost:\d+/"]

# Set copy-button to ignore python and bash prompts
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
github_repo = "ophyd-async"
github_user = "bluesky"
switcher_json = "https://blueskyproject.io/ophyd-async/switcher.json"
switcher_exists = requests.get(switcher_json).ok
if not switcher_exists:
    print(
        "*** Can't read version switcher, is GitHub pages enabled? \n"
        "    Once Docs CI job has successfully run once, set the "
        "Github pages source branch to be 'gh-pages' at:\n"
        f"    https://github.com/{github_user}/{github_repo}/settings/pages",
        file=sys.stderr,
    )

# Theme options for pydata_sphinx_theme
# We don't check switcher because there are 3 possible states for a repo:
# 1. New project, docs are not published so there is no switcher
# 2. Existing project with latest skeleton, switcher exists and works
# 3. Existing project with old skeleton that makes broken switcher,
#    switcher exists but is broken
# Point 3 makes checking switcher difficult, because the updated skeleton
# will fix the switcher at the end of the docs workflow, but never gets a chance
# to complete as the docs build warns and fails.
html_theme_options = {
    "use_edit_page_button": True,
    "github_url": f"https://github.com/{github_user}/{github_repo}",
    "icon_links": [
        {
            "name": "PyPI",
            "url": f"https://pypi.org/project/{project}",
            "icon": "fas fa-cube",
        },
    ],
    "switcher": {
        "json_url": switcher_json,
        "version_match": version,
    },
    "check_switcher": False,
    "navbar_end": ["theme-switcher", "icon-links", "version-switcher"],
    "external_links": [
        {
            "name": "Bluesky Project",
            "url": "https://blueskyproject.io",
        },
    ],
    "navigation_with_keys": False,
}

# A dictionary of values to pass into the template engine’s context for all pages
html_context = {
    "github_user": github_user,
    "github_repo": project,
    "github_version": version,
    "doc_path": "docs",
}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Logo
html_logo = "images/ophyd-async-logo.svg"
html_favicon = "images/ophyd-favicon.svg"

# If False and a module has the __all__ attribute set, autosummary documents
# every member listed in __all__ and no others. Default is True
autosummary_ignore_module_all = False

# Turn on sphinx.ext.autosummary
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Look for signatures in the first line of the docstring (used for C functions)
autodoc_docstring_signature = True

# numpydoc config
numpydoc_show_class_members = False

# Don't show config summary as it's not relevant
autodoc_pydantic_model_show_config_summary = False

# Show the fields in source order
autodoc_pydantic_model_summary_list_order = "bysource"

# Where to put Ipython savefigs
ipython_savefig_dir = "../build/savefig"



from pathlib import Path
from unittest.mock import ANY, MagicMock

import pytest

from ophyd_async.core import DeviceCollector, get_mock_put, set_mock_value
from ophyd_async.epics.eiger._odin_io import Odin, OdinWriter, Writing  # noqa: PLC2701

OdinDriverAndWriter = tuple[Odin, OdinWriter]


@pytest.fixture
def odin_driver_and_writer(RE) -> OdinDriverAndWriter:
    with DeviceCollector(mock=True):
        driver = Odin("")
        writer = OdinWriter(MagicMock(), lambda: "odin", driver)
    return driver, writer


async def test_when_open_called_then_file_correctly_set(
    odin_driver_and_writer: OdinDriverAndWriter, tmp_path: Path
):
    driver, writer = odin_driver_and_writer
    path_info = writer._path_provider.return_value
    expected_filename = "filename.h5"
    path_info.directory_path = tmp_path
    path_info.filename = expected_filename

    await writer.open()

    get_mock_put(driver.file_path).assert_called_once_with(str(tmp_path), wait=ANY)
    get_mock_put(driver.file_name).assert_called_once_with(expected_filename, wait=ANY)


async def test_when_open_called_then_all_expected_signals_set(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    await writer.open()

    get_mock_put(driver.data_type).assert_called_once_with("uint16", wait=ANY)
    get_mock_put(driver.num_to_capture).assert_called_once_with(0, wait=ANY)

    get_mock_put(driver.capture).assert_called_once_with(Writing.ON, wait=ANY)


async def test_given_data_shape_set_when_open_called_then_describe_has_correct_shape(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    set_mock_value(driver.image_width, 1024)
    set_mock_value(driver.image_height, 768)
    description = await writer.open()
    assert description["data"]["shape"] == [768, 1024]


async def test_when_closed_then_data_capture_turned_off(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    await writer.close()
    get_mock_put(driver.capture).assert_called_once_with(Writing.OFF, wait=ANY)



import asyncio
import time
from enum import Enum, IntEnum
from typing import Annotated as A

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import numpy as np
import pytest
from bluesky import RunEngine

import tango
from ophyd_async.core import Array1D, DeviceCollector, SignalRW, T
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.tango import TangoReadable, get_python_type
from ophyd_async.tango.demo import (
    DemoCounter,
    DemoMover,
    TangoDetector,
)
from tango import (
    AttrDataFormat,
    AttrQuality,
    AttrWriteType,
    CmdArgType,
    DevState,
)
from tango.asyncio import DeviceProxy as AsyncDeviceProxy
from tango.asyncio_executor import set_global_executor
from tango.server import Device, attribute, command
from tango.test_context import MultiDeviceTestContext
from tango.test_utils import assert_close


class TestEnum(IntEnum):
    __test__ = False
    A = 0
    B = 1


# --------------------------------------------------------------------
#               fixtures to run Echo device
# --------------------------------------------------------------------

TESTED_FEATURES = ["array", "limitedvalue", "justvalue"]


# --------------------------------------------------------------------
class TestDevice(Device):
    __test__ = False

    _array = [[1, 2, 3], [4, 5, 6]]

    _justvalue = 5
    _writeonly = 6
    _readonly = 7
    _slow_attribute = 1.0

    _floatvalue = 1.0

    _readback = 1.0
    _setpoint = 1.0

    _label = "Test Device"

    _limitedvalue = 3

    @attribute(dtype=float, access=AttrWriteType.READ)
    def readback(self):
        return self._readback

    @attribute(dtype=float, access=AttrWriteType.WRITE)
    def setpoint(self):
        return self._setpoint

    def write_setpoint(self, value: float):
        self._setpoint = value
        self._readback = value

    @attribute(dtype=str, access=AttrWriteType.READ_WRITE)
    def label(self):
        return self._label

    def write_label(self, value: str):
        self._label = value

    @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
    def floatvalue(self):
        return self._floatvalue

    def write_floatvalue(self, value: float):
        self._floatvalue = value

    @attribute(dtype=int, access=AttrWriteType.READ_WRITE, polling_period=100)
    def justvalue(self):
        return self._justvalue

    def write_justvalue(self, value: int):
        self._justvalue = value

    @attribute(dtype=int, access=AttrWriteType.WRITE, polling_period=100)
    def writeonly(self):
        return self._writeonly

    def write_writeonly(self, value: int):
        self._writeonly = value

    @attribute(dtype=int, access=AttrWriteType.READ, polling_period=100)
    def readonly(self):
        return self._readonly

    @attribute(
        dtype=float,
        access=AttrWriteType.READ_WRITE,
        dformat=AttrDataFormat.IMAGE,
        max_dim_x=3,
        max_dim_y=2,
    )
    def array(self) -> list[list[float]]:
        return self._array

    def write_array(self, array: list[list[float]]):
        self._array = array

    @attribute(
        dtype=float,
        access=AttrWriteType.READ_WRITE,
        min_value=0,
        min_alarm=1,
        min_warning=2,
        max_warning=4,
        max_alarm=5,
        max_value=6,
    )
    def limitedvalue(self) -> float:
        return self._limitedvalue

    def write_limitedvalue(self, value: float):
        self._limitedvalue = value

    @attribute(dtype=float, access=AttrWriteType.WRITE)
    def slow_attribute(self) -> float:
        return self._slow_attribute

    def write_slow_attribute(self, value: float):
        time.sleep(0.2)
        self._slow_attribute = value

    @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
    def raise_exception_attr(self) -> float:
        raise

    def write_raise_exception_attr(self, value: float):
        raise

    @command
    def clear(self) -> str:
        # self.info_stream("Received clear command")
        return "Received clear command"

    @command
    def slow_command(self) -> str:
        time.sleep(0.2)
        return "Completed slow command"

    @command
    def echo(self, value: str) -> str:
        return value

    @command
    def raise_exception_cmd(self):
        raise


# --------------------------------------------------------------------
class TestTangoReadable(TangoReadable):
    __test__ = False
    justvalue: A[SignalRW[int], Format.HINTED_UNCACHED_SIGNAL]
    array: A[SignalRW[Array1D[np.float64]], Format.HINTED_UNCACHED_SIGNAL]
    limitedvalue: A[SignalRW[float], Format.HINTED_UNCACHED_SIGNAL]


# --------------------------------------------------------------------
async def describe_class(fqtrl):
    description = {}
    values = {}
    dev = await AsyncDeviceProxy(fqtrl)

    for name in TESTED_FEATURES:
        if name in dev.get_attribute_list():
            attr_conf = await dev.get_attribute_config(name)
            attr_value = await dev.read_attribute(name)
            value = attr_value.value
            _, _, descr = get_python_type(attr_conf.data_type)
            max_x = attr_conf.max_dim_x
            max_y = attr_conf.max_dim_y
            if attr_conf.data_format == AttrDataFormat.SCALAR:
                is_array = False
                shape = []
            elif attr_conf.data_format == AttrDataFormat.SPECTRUM:
                is_array = True
                shape = [max_x]
            else:
                is_array = True
                shape = [max_y, max_x]

        elif name in dev.get_command_list():
            cmd_conf = await dev.get_command_config(name)
            _, _, descr = get_python_type(
                cmd_conf.in_type
                if cmd_conf.in_type != CmdArgType.DevVoid
                else cmd_conf.out_type
            )
            is_array = False
            shape = []
            value = getattr(dev, name)()

        else:
            raise RuntimeError(
                f"Cannot find {name} in attributes/commands (pipes are not supported!)"
            )

        description[f"test_device-{name}"] = {
            "source": f"{fqtrl}/{name}",  # type: ignore
            "dtype": "array" if is_array else descr,
            "shape": shape,
        }

        values[f"test_device-{name}"] = {
            "value": value,
            "timestamp": pytest.approx(time.time()),
            "alarm_severity": AttrQuality.ATTR_VALID,
        }

    return values, description


# --------------------------------------------------------------------
def get_test_descriptor(python_type: type[T], value: T, is_cmd: bool) -> dict:
    if python_type in [bool, int]:
        return {"dtype": "integer", "shape": []}
    if python_type in [float]:
        return {"dtype": "number", "shape": []}
    if python_type in [str]:
        return {"dtype": "string", "shape": []}
    if issubclass(python_type, DevState):
        return {"dtype": "string", "shape": [], "choices": list(DevState.names.keys())}
    if issubclass(python_type, Enum):
        return {
            "dtype": "string",
            "shape": [],
            "choices": [] if is_cmd else [member.name for member in python_type],
        }

    return {
        "dtype": "array",
        "shape": [np.Inf] if is_cmd else list(np.array(value).shape),
    }


# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def tango_test_device():
    with MultiDeviceTestContext(
        [{"class": TestDevice, "devices": [{"name": "test/device/1"}]}], process=True
    ) as context:
        yield context.get_device_access("test/device/1")


# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def demo_test_context():
    content = (
        {
            "class": DemoMover,
            "devices": [{"name": "demo/motor/1"}],
        },
        {
            "class": DemoCounter,
            "devices": [{"name": "demo/counter/1"}, {"name": "demo/counter/2"}],
        },
    )
    yield MultiDeviceTestContext(content, process=True)


# --------------------------------------------------------------------
@pytest.fixture(autouse=True)
def reset_tango_asyncio():
    set_global_executor(None)


# --------------------------------------------------------------------
def compare_values(expected, received):
    assert set(expected.keys()) == set(received.keys())
    for k, v in expected.items():
        for _k, _v in v.items():
            assert_close(_v, received[k][_k])


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_connect(tango_test_device):
    values, description = await describe_class(tango_test_device)

    async with DeviceCollector():
        test_device = TestTangoReadable(tango_test_device)

    assert test_device.name == "test_device"
    assert description == await test_device.describe()
    compare_values(values, await test_device.read())


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_set_trl(tango_test_device):
    values, description = await describe_class(tango_test_device)
    test_device = TestTangoReadable(name="test_device")

    test_device._connector.trl = tango_test_device
    await test_device.connect()

    assert test_device.name == "test_device"
    assert description == await test_device.describe()
    compare_values(values, await test_device.read())


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("proxy", [True, False, None])
async def test_connect_proxy(tango_test_device, proxy: bool | None):
    if proxy is None:
        test_device = TestTangoReadable(trl=tango_test_device)
        test_device.proxy = None
        await test_device.connect()
        assert isinstance(test_device._connector.proxy, tango._tango.DeviceProxy)
    elif proxy:
        proxy = await AsyncDeviceProxy(tango_test_device)
        test_device = TestTangoReadable(device_proxy=proxy)
        await test_device.connect()
        assert isinstance(test_device._connector.proxy, tango._tango.DeviceProxy)
    else:
        proxy = None
        test_device = TestTangoReadable(device_proxy=proxy)
        assert test_device


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_with_bluesky(tango_test_device):
    # now let's do some bluesky stuff
    RE = RunEngine()
    with DeviceCollector():
        device = TestTangoReadable(tango_test_device)
    RE(bp.count([device]))


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_demo(demo_test_context):
    with demo_test_context:
        detector = TangoDetector(
            name="detector",
            mover_trl="demo/motor/1",
            counter_trls=["demo/counter/1", "demo/counter/2"],
        )
        await detector.connect()
        await detector.trigger()
        await detector.mover.velocity.set(0.5)

        RE = RunEngine()

        RE(bps.read(detector))
        RE(bps.mv(detector, 0))
        RE(bp.count(list(detector.counters.values())))

        set_status = detector.set(1.0)
        await asyncio.sleep(1.0)
        stop_status = detector.stop()
        await set_status
        await stop_status
        assert all([set_status.done, stop_status.done])
        assert all([set_status.success, stop_status.success])



from pydantic import Field

from ophyd_async.core import AsyncStatus, PathProvider, StandardDetector, TriggerInfo

from ._eiger_controller import EigerController
from ._eiger_io import EigerDriverIO
from ._odin_io import Odin, OdinWriter


class EigerTriggerInfo(TriggerInfo):
    energy_ev: float = Field(gt=0)


class EigerDetector(StandardDetector):
    """
    Ophyd-async implementation of an Eiger Detector.
    """

    _controller: EigerController
    _writer: Odin

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix="-EA-EIGER-01:",
        hdf_suffix="-EA-ODIN-01:",
        name="",
    ):
        self.drv = EigerDriverIO(prefix + drv_suffix)
        self.odin = Odin(prefix + hdf_suffix + "FP:")

        super().__init__(
            EigerController(self.drv),
            OdinWriter(path_provider, lambda: self.name, self.odin),
            name=name,
        )

    @AsyncStatus.wrap
    async def prepare(self, value: EigerTriggerInfo) -> None:  # type: ignore
        await self._controller.set_energy(value.energy_ev)
        await super().prepare(value)



import asyncio

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adcore


class SimController(DetectorController):
    def __init__(
        self,
        driver: adcore.ADBaseIO,
        good_states: frozenset[adcore.DetectorState] = adcore.DEFAULT_GOOD_STATES,
    ) -> None:
        self.driver = driver
        self.good_states = good_states
        self.frame_timeout: float
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.002

    async def prepare(self, trigger_info: TriggerInfo):
        assert (
            trigger_info.trigger == DetectorTrigger.internal
        ), "fly scanning (i.e. external triggering) is not supported for this device"
        self.frame_timeout = (
            DEFAULT_TIMEOUT + await self.driver.acquire_time.get_value()
        )
        await asyncio.gather(
            self.driver.num_images.set(trigger_info.total_number_of_triggers),
            self.driver.image_mode.set(adcore.ImageMode.multiple),
        )

    async def arm(self):
        self._arm_status = await adcore.start_acquiring_driver_and_ensure_status(
            self.driver, good_states=self.good_states, timeout=self.frame_timeout
        )

    async def wait_for_idle(self):
        if self._arm_status:
            await self._arm_status

    async def disarm(self):
        # We can't use caput callback as we already used it in arm() and we can't have
        # 2 or they will deadlock
        await adcore.stop_busy_record(self.driver.acquire, False, timeout=1)



import asyncio
import time
from unittest.mock import AsyncMock, MagicMock, Mock, call

import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    AsyncStatus,
    DeviceCollector,
    callback_on_mock_put,
    mock_puts_blocked,
    observe_value,
    set_mock_put_proceeds,
    set_mock_value,
    soft_signal_rw,
)
from ophyd_async.epics import motor


async def wait_for_wakeups(max_yields=10):
    loop = asyncio.get_event_loop()
    # If anything has called loop.call_soon or is scheduled a wakeup
    # then let it run
    for _ in range(max_yields):
        await asyncio.sleep(0)
        if not loop._ready:
            return
    raise RuntimeError(f"Tasks still scheduling wakeups after {max_yields} yields")


@pytest.fixture
async def sim_motor():
    async with DeviceCollector(mock=True):
        sim_motor = motor.Motor("BLxxI-MO-TABLE-01:X", name="sim_motor")

    set_mock_value(sim_motor.motor_egu, "mm")
    set_mock_value(sim_motor.precision, 3)
    set_mock_value(sim_motor.velocity, 1)
    yield sim_motor


async def wait_for_eq(item, attribute, comparison, timeout):
    timeout_time = time.monotonic() + timeout
    while getattr(item, attribute) != comparison:
        await wait_for_wakeups()
        if time.monotonic() > timeout_time:
            raise TimeoutError


async def test_motor_moving_well(sim_motor: motor.Motor) -> None:
    set_mock_put_proceeds(sim_motor.user_setpoint, False)
    s = sim_motor.set(0.55)
    watcher = Mock()
    s.watch(watcher)
    done = Mock()
    s.add_callback(done)
    await wait_for_wakeups()
    await wait_for_eq(watcher, "call_count", 1, 1)
    assert watcher.call_args == call(
        name="sim_motor",
        current=0.0,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.05),
    )
    watcher.reset_mock()
    assert 0.55 == await sim_motor.user_setpoint.get_value()
    assert not s.done
    await asyncio.sleep(0.1)
    set_mock_value(sim_motor.user_readback, 0.1)
    await wait_for_eq(watcher, "call_count", 1, 1)
    assert watcher.call_count == 1
    assert watcher.call_args == call(
        name="sim_motor",
        current=0.1,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.1, abs=0.05),
    )
    set_mock_value(sim_motor.motor_done_move, True)
    set_mock_value(sim_motor.user_readback, 0.55)
    set_mock_put_proceeds(sim_motor.user_setpoint, True)
    await wait_for_wakeups()
    await wait_for_eq(s, "done", True, 1)
    done.assert_called_once_with(s)


async def test_motor_moving_well_2(sim_motor: motor.Motor) -> None:
    set_mock_put_proceeds(sim_motor.user_setpoint, False)
    s = sim_motor.set(0.55)
    watcher = Mock()
    s.watch(watcher)
    done = Mock()
    s.add_callback(done)
    await wait_for_wakeups()
    assert watcher.call_count == 1
    assert watcher.call_args == call(
        name="sim_motor",
        current=0.0,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.2),
    )
    watcher.reset_mock()
    assert 0.55 == await sim_motor.user_setpoint.get_value()
    assert not s.done
    await asyncio.sleep(0.1)
    set_mock_value(sim_motor.user_readback, 0.1)
    await asyncio.sleep(0.1)
    assert watcher.call_count == 1
    assert watcher.call_args == call(
        name="sim_motor",
        current=0.1,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.1, abs=0.2),
    )
    set_mock_put_proceeds(sim_motor.user_setpoint, True)
    await wait_for_wakeups()
    assert s.done
    done.assert_called_once_with(s)


async def test_motor_move_timeout(sim_motor: motor.Motor):
    class MyTimeout(Exception):
        pass

    def do_timeout(value, wait):
        # Raise custom exception to be clear it bubbles up
        raise MyTimeout()

    callback_on_mock_put(sim_motor.user_setpoint, do_timeout)
    s = sim_motor.set(0.3)
    watcher = Mock()
    s.watch(watcher)
    with pytest.raises(MyTimeout):
        await s
    watcher.assert_called_once_with(
        name="sim_motor",
        current=0.0,
        initial=0.0,
        target=0.3,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.05),
    )


async def test_motor_moving_stopped(sim_motor: motor.Motor):
    set_mock_value(sim_motor.motor_done_move, False)
    set_mock_put_proceeds(sim_motor.user_setpoint, False)
    s = sim_motor.set(1.5)
    s.add_callback(Mock())
    await asyncio.sleep(0.2)
    assert not s.done
    await sim_motor.stop()
    set_mock_put_proceeds(sim_motor.user_setpoint, True)
    await wait_for_wakeups()
    assert s.done
    assert s.success is False


async def test_read_motor(sim_motor: motor.Motor):
    await sim_motor.stage()
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.0
    assert (await sim_motor.read_configuration())["sim_motor-velocity"]["value"] == 1
    assert (await sim_motor.describe_configuration())["sim_motor-motor_egu"][
        "shape"
    ] == []
    set_mock_value(sim_motor.user_readback, 0.5)
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.5
    await sim_motor.unstage()
    # Check we can still read and describe when not staged
    set_mock_value(sim_motor.user_readback, 0.1)
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.1
    assert await sim_motor.describe()


async def test_set_velocity(sim_motor: motor.Motor) -> None:
    v = sim_motor.velocity
    q: asyncio.Queue[dict[str, Reading]] = asyncio.Queue()
    v.subscribe(q.put_nowait)
    assert (await q.get())["sim_motor-velocity"]["value"] == 1.0
    await v.set(2.0)
    assert (await q.get())["sim_motor-velocity"]["value"] == 2.0
    v.clear_sub(q.put_nowait)
    await v.set(3.0)
    assert (await v.read())["sim_motor-velocity"]["value"] == 3.0
    assert q.empty()


async def test_prepare_velocity_errors(sim_motor: motor.Motor):
    set_mock_value(sim_motor.max_velocity, 10)
    with pytest.raises(motor.MotorLimitsException):
        fly_info = motor.FlyMotorInfo(
            start_position=-10, end_position=0, time_for_move=0.9
        )
        await sim_motor._prepare_velocity(
            fly_info.start_position,
            fly_info.end_position,
            fly_info.time_for_move,
        )


async def test_valid_prepare_velocity(sim_motor: motor.Motor):
    set_mock_value(sim_motor.max_velocity, 10)
    fly_info = motor.FlyMotorInfo(start_position=-10, end_position=0, time_for_move=1)
    await sim_motor._prepare_velocity(
        fly_info.start_position,
        fly_info.end_position,
        fly_info.time_for_move,
    )
    assert (
        await sim_motor.velocity.get_value() == await sim_motor.max_velocity.get_value()
    )


@pytest.mark.parametrize(
    "acceleration_time, velocity, start_position, end_position, upper_limit,\
    lower_limit",
    [
        (1, 10, 0, 10, 30, -4.999),  # Goes below lower_limit, +ve direction
        (1, 10, 0, 10, 14.99, -10),  # Goes above upper_limit, +ve direction
        (1, -10, 10, 0, -30, -9.999),  # Goes below lower_limit, -ve direction
        (1, -10, 10, 0, 14.99, -10),  # Goes above upper_limit, -ve direction
    ],
)
async def test_prepare_motor_path_errors(
    sim_motor: motor.Motor,
    acceleration_time,
    velocity,
    start_position,
    end_position,
    upper_limit,
    lower_limit,
):
    set_mock_value(sim_motor.acceleration_time, acceleration_time)
    set_mock_value(sim_motor.low_limit_travel, lower_limit)
    set_mock_value(sim_motor.high_limit_travel, upper_limit)
    with pytest.raises(motor.MotorLimitsException):
        await sim_motor._prepare_motor_path(velocity, start_position, end_position)


async def test_prepare_motor_path(sim_motor: motor.Motor):
    set_mock_value(sim_motor.acceleration_time, 1)
    set_mock_value(sim_motor.low_limit_travel, -10.01)
    set_mock_value(sim_motor.high_limit_travel, 20.01)
    fly_info = motor.FlyMotorInfo(
        start_position=0,
        end_position=10,
        time_for_move=1,
    )
    assert (
        await sim_motor._prepare_motor_path(
            10, fly_info.start_position, fly_info.end_position
        )
        == -5
    )
    assert sim_motor._fly_completed_position == 15


@pytest.mark.parametrize(
    "expected_velocity, target_position",
    [
        (10, -10),
        (8, 8),
    ],
)
async def test_prepare(
    sim_motor: motor.Motor, target_position: float, expected_velocity: float
):
    set_mock_value(sim_motor.acceleration_time, 1)
    set_mock_value(sim_motor.low_limit_travel, -10)
    set_mock_value(sim_motor.high_limit_travel, 20)
    set_mock_value(sim_motor.max_velocity, 10)
    fake_set_signal = soft_signal_rw(float)
    await fake_set_signal.connect()

    async def wait_for_set(_):
        async for value in observe_value(fake_set_signal, timeout=1):
            if value == target_position:
                break

    sim_motor.set = AsyncMock(side_effect=wait_for_set)

    async def do_set(status: AsyncStatus):
        assert not status.done
        await fake_set_signal.set(target_position)

    async def wait_for_status(status: AsyncStatus):
        await status

    status = sim_motor.prepare(
        motor.FlyMotorInfo(
            start_position=0,
            end_position=target_position,
            time_for_move=1,
        )
    )
    # Test that prepare is not marked as complete until correct position is reached
    await asyncio.gather(do_set(status), wait_for_status(status))
    assert await sim_motor.velocity.get_value() == expected_velocity
    assert status.done


async def test_kickoff(sim_motor: motor.Motor):
    sim_motor.set = MagicMock()
    with pytest.raises(AssertionError):
        await sim_motor.kickoff()
    with pytest.raises(AssertionError):
        await sim_motor.kickoff()
    sim_motor._fly_completed_position = 20
    await sim_motor.kickoff()
    sim_motor.set.assert_called_once_with(20, timeout=CALCULATE_TIMEOUT)


async def test_complete(sim_motor: motor.Motor) -> None:
    with pytest.raises(AssertionError):
        sim_motor.complete()
    sim_motor._fly_status = sim_motor.set(20)
    assert not sim_motor._fly_status.done
    await sim_motor.complete()
    assert sim_motor._fly_status.done


async def test_locatable(sim_motor: motor.Motor) -> None:
    callback_on_mock_put(
        sim_motor.user_setpoint,
        lambda x, *_, **__: set_mock_value(sim_motor.user_readback, x),
    )
    assert (await sim_motor.locate())["readback"] == 0
    async with mock_puts_blocked(sim_motor.user_setpoint):
        move_status = sim_motor.set(10)
        assert (await sim_motor.locate())["readback"] == 0
    await move_status
    assert (await sim_motor.locate())["readback"] == 10
    assert (await sim_motor.locate())["setpoint"] == 10



import asyncio
import os
import pprint
import subprocess
import sys
import time
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from bluesky.run_engine import RunEngine, TransitionError
from pytest import FixtureRequest

from ophyd_async.core import (
    DetectorTrigger,
    FilenameProvider,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
)

PANDA_RECORD = str(Path(__file__).parent / "fastcs" / "panda" / "db" / "panda.db")
INCOMPLETE_BLOCK_RECORD = str(
    Path(__file__).parent / "fastcs" / "panda" / "db" / "incomplete_block_panda.db"
)
INCOMPLETE_RECORD = str(
    Path(__file__).parent / "fastcs" / "panda" / "db" / "incomplete_panda.db"
)
EXTRA_BLOCKS_RECORD = str(
    Path(__file__).parent / "fastcs" / "panda" / "db" / "extra_blocks_panda.db"
)

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[Any]):
        if call.excinfo is not None:
            raise call.excinfo.value
        else:
            raise RuntimeError(
                f"{call} has no exception data, an unknown error has occurred"
            )

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[Any]):
        raise excinfo.value


# Autouse fixture that will set all EPICS networking env vars to use lo interface
# to avoid false failures caused by things like firewalls blocking EPICS traffic.
@pytest.fixture(scope="session", autouse=True)
def configure_epics_environment():
    os.environ["EPICS_CAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_AUTO_ADDR_LIST"] = "NO"
    os.environ["EPICS_CA_AUTO_BEACON_ADDR_LIST"] = "NO"

    os.environ["EPICS_PVAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_AUTO_BEACON_ADDR_LIST"] = "NO"
    os.environ["EPICS_PVA_AUTO_ADDR_LIST"] = "NO"


_ALLOWED_PYTEST_TASKS = {"async_finalizer", "async_setup", "async_teardown"}


def _error_and_kill_pending_tasks(
    loop: asyncio.AbstractEventLoop, test_name: str, test_passed: bool
) -> set[asyncio.Task]:
    """Cancels pending tasks in the event loop for a test. Raises an exception if
    the test hasn't already.

    Args:
        loop: The event loop to check for pending tasks.
        test_name: The name of the test.
        test_passed: Indicates whether the test passed.

    Returns:
        set[asyncio.Task]: The set of unfinished tasks that were cancelled.

    Raises:
        RuntimeError: If there are unfinished tasks and the test didn't fail.
    """
    unfinished_tasks = {
        task
        for task in asyncio.all_tasks(loop)
        if task.get_coro().__name__ not in _ALLOWED_PYTEST_TASKS and not task.done()
    }
    for task in unfinished_tasks:
        task.cancel()

    # We only raise an exception here if the test didn't fail anyway.
    # If it did then it makes sense that there's some tasks we need to cancel,
    # but an exception will already have been raised.
    if unfinished_tasks and test_passed:
        raise RuntimeError(
            f"Not all tasks closed during test {test_name}:\n"
            f"{pprint.pformat(unfinished_tasks, width=88)}"
        )

    return unfinished_tasks


@pytest.fixture(autouse=True, scope="function")
def fail_test_on_unclosed_tasks(request: FixtureRequest):
    """
    Used on every test to ensure failure if there are pending tasks
    by the end of the test.
    """

    fail_count = request.session.testsfailed
    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    request.addfinalizer(
        lambda: _error_and_kill_pending_tasks(
            loop, request.node.name, request.session.testsfailed == fail_count
        )
    )


@pytest.fixture(scope="function")
def RE(request: FixtureRequest):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=True, loop=loop)
    fail_count = request.session.testsfailed

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass

        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()

        try:
            _error_and_kill_pending_tasks(
                loop, request.node.name, request.session.testsfailed == fail_count
            )
        finally:
            loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="module", params=["pva"])
def panda_pva():
    processes = [
        subprocess.Popen(
            [
                sys.executable,
                "-m",
                "epicscorelibs.ioc",
                "-m",
                macros,
                "-d",
                PANDA_RECORD,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        for macros in [
            "INCLUDE_EXTRA_BLOCK=,INCLUDE_EXTRA_SIGNAL=",
            "EXCLUDE_WIDTH=#,IOC_NAME=PANDAQSRVIB",
            "EXCLUDE_PCAP=#,IOC_NAME=PANDAQSRVI",
        ]
    ]
    time.sleep(2)

    for p in processes:
        assert not p.poll(), p.stdout.read()

    yield processes

    for p in processes:
        p.terminate()
        p.communicate()


@pytest.fixture
async def normal_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)

    return inner_coroutine


@pytest.fixture
async def failing_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)
        raise ValueError()

    return inner_coroutine


@pytest.fixture
def static_filename_provider():
    return StaticFilenameProvider("ophyd_async_tests")


@pytest.fixture
def static_path_provider_factory(tmp_path: Path):
    def create_static_dir_provider_given_fp(fp: FilenameProvider):
        return StaticPathProvider(fp, tmp_path)

    return create_static_dir_provider_given_fp


@pytest.fixture
def static_path_provider(
    static_path_provider_factory: Callable,
    static_filename_provider: FilenameProvider,
):
    return static_path_provider_factory(static_filename_provider)


@pytest.fixture
def one_shot_trigger_info() -> TriggerInfo:
    return TriggerInfo(
        frame_timeout=None,
        number_of_triggers=1,
        trigger=DetectorTrigger.internal,
        deadtime=None,
        livetime=None,
    )


def pytest_collection_modifyitems(config, items):
    tango_dir = "tests/tango"

    for item in items:
        if tango_dir in str(item.fspath):
            item.add_marker(pytest.mark.forked)



import os
import uuid
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import Protocol


@dataclass
class PathInfo:
    """
    Information about where and how to write a file.

    :param directory_path: Directory into which files should be written
    :param filename: Base filename to use generated by FilenameProvider, w/o extension
    :param create_dir_depth: Optional depth of directories to create if they do not
                             exist
    """

    directory_path: Path
    filename: str
    create_dir_depth: int = 0


class FilenameProvider(Protocol):
    @abstractmethod
    def __call__(self, device_name: str | None = None) -> str:
        """Get a filename to use for output data, w/o extension"""


class PathProvider(Protocol):
    @abstractmethod
    def __call__(self, device_name: str | None = None) -> PathInfo:
        """Get the current directory to write files into"""


class StaticFilenameProvider(FilenameProvider):
    def __init__(self, filename: str):
        self._static_filename = filename

    def __call__(self, device_name: str | None = None) -> str:
        return self._static_filename


class UUIDFilenameProvider(FilenameProvider):
    def __init__(
        self,
        uuid_call_func: Callable = uuid.uuid4,
        uuid_call_args: list | None = None,
    ):
        self._uuid_call_func = uuid_call_func
        self._uuid_call_args = uuid_call_args or []

    def __call__(self, device_name: str | None = None) -> str:
        if (
            self._uuid_call_func in [uuid.uuid3, uuid.uuid5]
            and len(self._uuid_call_args) < 2
        ):
            raise ValueError(
                f"To use {self._uuid_call_func} to generate UUID filenames,"
                " UUID namespace and name must be passed as args!"
            )

        uuid_str = self._uuid_call_func(*self._uuid_call_args)
        return f"{uuid_str}"


class AutoIncrementFilenameProvider(FilenameProvider):
    def __init__(
        self,
        base_filename: str = "",
        max_digits: int = 5,
        starting_value: int = 0,
        increment: int = 1,
        inc_delimeter: str = "_",
    ):
        self._base_filename = base_filename
        self._max_digits = max_digits
        self._current_value = starting_value
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, device_name: str | None = None) -> str:
        if len(str(self._current_value)) > self._max_digits:
            raise ValueError(
                f"Auto incrementing filename counter \
                  exceeded maximum of {self._max_digits} digits!"
            )

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        filename = f"{self._base_filename}{self._inc_delimeter}{padded_counter}"

        self._current_value += self._increment
        return filename


class StaticPathProvider(PathProvider):
    def __init__(
        self,
        filename_provider: FilenameProvider,
        directory_path: Path,
        create_dir_depth: int = 0,
    ) -> None:
        self._filename_provider = filename_provider
        self._directory_path = directory_path
        self._create_dir_depth = create_dir_depth

    def __call__(self, device_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(device_name)

        return PathInfo(
            directory_path=self._directory_path,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class AutoIncrementingPathProvider(PathProvider):
    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: Path,
        create_dir_depth: int = 0,
        max_digits: int = 5,
        starting_value: int = 0,
        num_calls_per_inc: int = 1,
        increment: int = 1,
        inc_delimeter: str = "_",
        base_name: str | None = None,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = base_directory_path
        self._create_dir_depth = create_dir_depth
        self._base_name = base_name
        self._starting_value = starting_value
        self._current_value = starting_value
        self._num_calls_per_inc = num_calls_per_inc
        self._inc_counter = 0
        self._max_digits = max_digits
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, device_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(device_name)

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        auto_inc_dir_name = str(padded_counter)
        if self._base_name is not None:
            auto_inc_dir_name = (
                f"{self._base_name}{self._inc_delimeter}{padded_counter}"
            )
        elif device_name is not None:
            auto_inc_dir_name = f"{device_name}{self._inc_delimeter}{padded_counter}"

        self._inc_counter += 1
        if self._inc_counter == self._num_calls_per_inc:
            self._inc_counter = 0
            self._current_value += self._increment

        return PathInfo(
            directory_path=self._base_directory_path / auto_inc_dir_name,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class YMDPathProvider(PathProvider):
    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: Path,
        create_dir_depth: int = -3,  # Default to -3 to create YMD dirs
        device_name_as_base_dir: bool = False,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = Path(base_directory_path)
        self._create_dir_depth = create_dir_depth
        self._device_name_as_base_dir = device_name_as_base_dir

    def __call__(self, device_name: str | None = None) -> PathInfo:
        sep = os.path.sep
        current_date = date.today().strftime(f"%Y{sep}%m{sep}%d")
        if device_name is None:
            ymd_dir_path = current_date
        elif self._device_name_as_base_dir:
            ymd_dir_path = os.path.join(
                current_date,
                device_name,
            )
        else:
            ymd_dir_path = os.path.join(
                device_name,
                current_date,
            )

        filename = self._filename_provider(device_name)
        return PathInfo(
            directory_path=self._base_directory_path / ymd_dir_path,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class NameProvider(Protocol):
    @abstractmethod
    def __call__(self) -> str:
        """Get the name to be used as a data_key in the descriptor document"""


class DatasetDescriber(Protocol):
    @abstractmethod
    async def np_datatype(self) -> str:
        """Represents the numpy datatype"""

    @abstractmethod
    async def shape(self) -> tuple[int, ...]:
        """Get the shape of the data collection"""



import asyncio
import textwrap
import time
from enum import Enum, IntEnum
from random import choice

import numpy as np
import numpy.typing as npt
import pytest
from bluesky.protocols import Reading
from test_base_device import TestDevice

from ophyd_async.core import SignalBackend, SignalR, SignalRW, SignalW, SignalX, T
from ophyd_async.tango import (
    TangoSignalBackend,
    tango_signal_r,
    tango_signal_rw,
    tango_signal_w,
    tango_signal_x,
)
from tango import AttrDataFormat, AttrWriteType, DevState
from tango.asyncio import DeviceProxy
from tango.asyncio_executor import set_global_executor
from tango.server import Device, attribute, command
from tango.test_context import MultiDeviceTestContext
from tango.test_utils import assert_close


def __tango_signal_auto(*args, **kwargs):
    raise RuntimeError("Fix this later")


# --------------------------------------------------------------------
"""
Since TangoTest does not support EchoMode, we create our own Device.

"""


class TestEnum(IntEnum):
    __test__ = False
    A = 0
    B = 1


BASE_TYPES_SET = (
    # type_name, tango_name,    py_type,    sample_values
    ("boolean", "DevBoolean", bool, (True, False)),
    ("short", "DevShort", int, (1, 2, 3, 4, 5)),
    ("ushort", "DevUShort", int, (1, 2, 3, 4, 5)),
    ("long", "DevLong", int, (1, 2, 3, 4, 5)),
    ("ulong", "DevULong", int, (1, 2, 3, 4, 5)),
    ("long64", "DevLong64", int, (1, 2, 3, 4, 5)),
    ("char", "DevUChar", int, (1, 2, 3, 4, 5)),
    ("float", "DevFloat", float, (1.1, 2.2, 3.3, 4.4, 5.5)),
    ("double", "DevDouble", float, (1.1, 2.2, 3.3, 4.4, 5.5)),
    ("string", "DevString", str, ("aaa", "bbb", "ccc")),
    ("state", "DevState", DevState, (DevState.ON, DevState.MOVING, DevState.ALARM)),
    ("enum", "DevEnum", TestEnum, (TestEnum.A, TestEnum.B)),
)

ATTRIBUTES_SET = []
COMMANDS_SET = []

for type_name, tango_type_name, py_type, values in BASE_TYPES_SET:
    ATTRIBUTES_SET.extend(
        [
            (
                f"{type_name}_scalar_attr",
                tango_type_name,
                AttrDataFormat.SCALAR,
                py_type,
                choice(values),
                choice(values),
            ),
            (
                f"{type_name}_spectrum_attr",
                tango_type_name,
                AttrDataFormat.SPECTRUM,
                npt.NDArray[py_type],
                [choice(values), choice(values), choice(values)],
                [choice(values), choice(values), choice(values)],
            ),
            (
                f"{type_name}_image_attr",
                tango_type_name,
                AttrDataFormat.IMAGE,
                npt.NDArray[py_type],
                [
                    [choice(values), choice(values), choice(values)],
                    [choice(values), choice(values), choice(values)],
                ],
                [
                    [choice(values), choice(values), choice(values)],
                    [choice(values), choice(values), choice(values)],
                ],
            ),
        ]
    )

    if tango_type_name == "DevUChar":
        continue
    else:
        COMMANDS_SET.append(
            (
                f"{type_name}_scalar_cmd",
                tango_type_name,
                AttrDataFormat.SCALAR,
                py_type,
                choice(values),
                choice(values),
            )
        )
        if tango_type_name in ["DevState", "DevEnum"]:
            continue
        else:
            COMMANDS_SET.append(
                (
                    f"{type_name}_spectrum_cmd",
                    tango_type_name,
                    AttrDataFormat.SPECTRUM,
                    npt.NDArray[py_type],
                    [choice(values), choice(values), choice(values)],
                    [choice(values), choice(values), choice(values)],
                )
            )


# --------------------------------------------------------------------
#               TestDevice
# --------------------------------------------------------------------
# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def tango_test_device():
    with MultiDeviceTestContext(
        [{"class": TestDevice, "devices": [{"name": "test/device/1"}]}], process=True
    ) as context:
        yield context.get_device_access("test/device/1")


# --------------------------------------------------------------------
#               Echo device
# --------------------------------------------------------------------
class EchoDevice(Device):
    attr_values = {}

    def initialize_dynamic_attributes(self):
        for name, typ, form, _, _, _ in ATTRIBUTES_SET:
            attr = attribute(
                name=name,
                dtype=typ,
                dformat=form,
                access=AttrWriteType.READ_WRITE,
                fget=self.read,
                fset=self.write,
                max_dim_x=3,
                max_dim_y=2,
                enum_labels=[member.name for member in TestEnum],
            )
            self.add_attribute(attr)
            self.set_change_event(name, True, False)

        for name, typ, form, _, _, _ in COMMANDS_SET:
            cmd = command(
                f=getattr(self, name),
                dtype_in=typ,
                dformat_in=form,
                dtype_out=typ,
                dformat_out=form,
            )
            self.add_command(cmd)

    def read(self, attr):
        attr.set_value(self.attr_values[attr.get_name()])

    def write(self, attr):
        new_value = attr.get_write_value()
        self.attr_values[attr.get_name()] = new_value
        self.push_change_event(attr.get_name(), new_value)

    echo_command_code = textwrap.dedent(
        """\
            def echo_command(self, arg):
                return arg
            """
    )

    for name, _, _, _, _, _ in COMMANDS_SET:
        exec(echo_command_code.replace("echo_command", name))


# --------------------------------------------------------------------
def assert_enum(initial_value, readout_value):
    if type(readout_value) in [list, tuple]:
        for _initial_value, _readout_value in zip(
            initial_value, readout_value, strict=False
        ):
            assert_enum(_initial_value, _readout_value)
    else:
        assert initial_value == readout_value


# --------------------------------------------------------------------
#               fixtures to run Echo device
# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def echo_device():
    with MultiDeviceTestContext(
        [{"class": EchoDevice, "devices": [{"name": "test/device/1"}]}], process=True
    ) as context:
        yield context.get_device_access("test/device/1")


# --------------------------------------------------------------------
@pytest.fixture(autouse=True)
def reset_tango_asyncio():
    set_global_executor(None)


# --------------------------------------------------------------------
#               helpers to run tests
# --------------------------------------------------------------------
def get_test_descriptor(python_type: type[T], value: T, is_cmd: bool) -> dict:
    if python_type in [bool, int]:
        return {"dtype": "integer", "shape": []}
    if python_type in [float]:
        return {"dtype": "number", "shape": []}
    if python_type in [str]:
        return {"dtype": "string", "shape": []}
    if issubclass(python_type, DevState):
        return {"dtype": "string", "shape": []}
    if issubclass(python_type, Enum):
        return {
            "dtype": "string",
            "shape": [],
        }

    return {
        "dtype": "array",
        "shape": [np.iinfo(np.int32).max] if is_cmd else list(np.array(value).shape),
    }


# --------------------------------------------------------------------
async def make_backend(
    typ: type | None,
    pv: str,
    connect: bool = True,
    allow_events: bool | None = True,
) -> TangoSignalBackend:
    backend = TangoSignalBackend(typ, pv, pv)
    backend.allow_events(allow_events)
    if connect:
        await backend.connect(1)
    return backend


# --------------------------------------------------------------------
async def prepare_device(echo_device: str, pv: str, put_value: T) -> None:
    proxy = await DeviceProxy(echo_device)
    setattr(proxy, pv, put_value)


# --------------------------------------------------------------------
class MonitorQueue:
    def __init__(self, backend: SignalBackend):
        self.updates: asyncio.Queue[Reading] = asyncio.Queue()
        self.backend = backend
        self.subscription = backend.set_callback(self.updates.put_nowait)

    async def assert_updates(self, expected_value):
        expected_reading = {
            "timestamp": pytest.approx(time.time(), rel=0.1),
            "alarm_severity": 0,
        }
        update_reading = dict(await asyncio.wait_for(self.updates.get(), timeout=5))
        update_value = update_reading.pop("value")
        assert_close(update_value, expected_value)
        backend_reading = dict(
            await asyncio.wait_for(self.backend.get_reading(), timeout=5)
        )
        backend_reading.pop("value")
        backend_value = await asyncio.wait_for(self.backend.get_value(), timeout=5)
        assert_close(backend_value, expected_value)
        assert update_reading == expected_reading == backend_reading

    def close(self):
        self.backend.set_callback(None)


# --------------------------------------------------------------------
async def assert_monitor_then_put(
    echo_device: str,
    pv: str,
    initial_value: T,
    put_value: T,
    descriptor: dict,
    datatype: type[T] | None = None,
):
    await prepare_device(echo_device, pv, initial_value)
    source = echo_device + "/" + pv
    backend = await make_backend(datatype, source, allow_events=True)
    # Make a monitor queue that will monitor for updates
    q = MonitorQueue(backend)
    try:
        assert dict(source=source, **descriptor) == await backend.get_datakey("")
        # Check initial value
        await q.assert_updates(initial_value)
        # Put to new value and check that
        await backend.put(put_value, wait=True)
        assert_close(put_value, await backend.get_setpoint())
        await q.assert_updates(put_value)
    finally:
        q.close()


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value",
    ATTRIBUTES_SET,
    ids=[x[0] for x in ATTRIBUTES_SET],
)
async def test_backend_get_put_monitor_attr(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
):
    try:
        # Set a timeout for the operation to prevent it from running indefinitely
        await asyncio.wait_for(
            assert_monitor_then_put(
                echo_device,
                pv,
                initial_value,
                put_value,
                get_test_descriptor(py_type, initial_value, False),
                py_type,
            ),
            timeout=100,  # Timeout in seconds
        )
    except asyncio.TimeoutError:
        pytest.fail("Test timed out")
    except Exception as e:
        pytest.fail(f"Test failed with exception: {e}")


# --------------------------------------------------------------------
async def assert_put_read(
    echo_device: str,
    pv: str,
    put_value: T,
    descriptor: dict,
    datatype: type[T] | None = None,
):
    source = echo_device + "/" + pv
    backend = await make_backend(datatype, source)
    # Make a monitor queue that will monitor for updates
    assert dict(source=source, **descriptor) == await backend.get_datakey("")
    # Put to new value and check that
    await backend.put(put_value, wait=True)

    expected_reading = {
        "timestamp": pytest.approx(time.time(), rel=0.1),
        "alarm_severity": 0,
    }

    assert_close(await backend.get_value(), put_value)

    get_reading = dict(await backend.get_reading())
    get_reading.pop("value")
    assert expected_reading == get_reading


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value",
    COMMANDS_SET,
    ids=[x[0] for x in COMMANDS_SET],
)
async def test_backend_get_put_monitor_cmd(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
):
    # With the given datatype, check we have the correct initial value and putting works
    descriptor = get_test_descriptor(py_type, initial_value, True)
    await assert_put_read(echo_device, pv, put_value, descriptor, py_type)
    # # With guessed datatype, check we can set it back to the initial value
    await assert_put_read(echo_device, pv, put_value, descriptor)
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    await asyncio.gather(*tasks)
    del echo_device


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value, use_proxy",
    [
        (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
            use_proxy,
        )
        for (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
        ) in ATTRIBUTES_SET
        for use_proxy in [True, False]
    ],
    ids=[f"{x[0]}_{use_proxy}" for x in ATTRIBUTES_SET for use_proxy in [True, False]],
)
async def test_tango_signal_r(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
    use_proxy: bool,
):
    await prepare_device(echo_device, pv, initial_value)
    source = echo_device + "/" + pv
    proxy = await DeviceProxy(echo_device) if use_proxy else None

    timeout = 0.2
    signal = tango_signal_r(
        datatype=py_type,
        read_trl=source,
        device_proxy=proxy,
        timeout=timeout,
        name="test_signal",
    )
    await signal.connect()
    reading = await signal.read()
    assert_close(reading["test_signal"]["value"], initial_value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value, use_proxy",
    [
        (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
            use_proxy,
        )
        for (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
        ) in ATTRIBUTES_SET
        for use_proxy in [True, False]
    ],
    ids=[f"{x[0]}_{use_proxy}" for x in ATTRIBUTES_SET for use_proxy in [True, False]],
)
async def test_tango_signal_w(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
    use_proxy: bool,
):
    await prepare_device(echo_device, pv, initial_value)
    source = echo_device + "/" + pv
    proxy = await DeviceProxy(echo_device) if use_proxy else None

    timeout = 0.2
    signal = tango_signal_w(
        datatype=py_type,
        write_trl=source,
        device_proxy=proxy,
        timeout=timeout,
        name="test_signal",
    )
    await signal.connect()
    status = signal.set(put_value, wait=True, timeout=timeout)
    await status
    assert status.done is True and status.success is True

    status = signal.set(put_value, wait=False, timeout=timeout)
    await status
    assert status.done is True and status.success is True

    status = signal.set(put_value, wait=True)
    await status
    assert status.done is True and status.success is True

    status = signal.set(put_value, wait=False)
    await status
    assert status.done is True and status.success is True


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value, use_proxy",
    [
        (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
            use_proxy,
        )
        for (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
        ) in ATTRIBUTES_SET
        for use_proxy in [True, False]
    ],
    ids=[f"{x[0]}_{use_proxy}" for x in ATTRIBUTES_SET for use_proxy in [True, False]],
)
async def test_tango_signal_rw(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
    use_proxy: bool,
):
    await prepare_device(echo_device, pv, initial_value)
    source = echo_device + "/" + pv
    proxy = await DeviceProxy(echo_device) if use_proxy else None

    timeout = 0.2
    signal = tango_signal_rw(
        datatype=py_type,
        read_trl=source,
        write_trl=source,
        device_proxy=proxy,
        timeout=timeout,
        name="test_signal",
    )
    await signal.connect()
    reading = await signal.read()
    assert_close(reading["test_signal"]["value"], initial_value)
    await signal.set(put_value)
    location = await signal.locate()
    assert_close(location["setpoint"], put_value)
    assert_close(location["readback"], put_value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("use_proxy", [True, False])
async def test_tango_signal_x(tango_test_device: str, use_proxy: bool):
    proxy = await DeviceProxy(tango_test_device) if use_proxy else None
    timeout = 0.2
    signal = tango_signal_x(
        write_trl=tango_test_device + "/" + "clear",
        device_proxy=proxy,
        timeout=timeout,
        name="test_signal",
    )
    await signal.connect()
    status = signal.trigger()
    await status
    assert status.done is True and status.success is True


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.skip("Not sure if we need tango_signal_auto")
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value, use_proxy",
    [
        (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
            use_proxy,
        )
        for (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
        ) in ATTRIBUTES_SET
        for use_proxy in [True, False]
    ],
    ids=[f"{x[0]}_{use_proxy}" for x in ATTRIBUTES_SET for use_proxy in [True, False]],
)
async def test_tango_signal_auto_attrs(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
    use_proxy: bool,
):
    await prepare_device(echo_device, pv, initial_value)
    source = echo_device + "/" + pv
    proxy = await DeviceProxy(echo_device) if use_proxy else None
    timeout = 0.2

    async def _test_signal(dtype, proxy):
        signal = await __tango_signal_auto(
            datatype=dtype,
            trl=source,
            device_proxy=proxy,
            timeout=timeout,
            name="test_signal",
        )
        assert type(signal) is SignalRW
        await signal.connect()
        reading = await signal.read()
        value = reading["test_signal"]["value"]
        if isinstance(value, np.ndarray):
            value = value.tolist()
        assert_close(value, initial_value)

        await signal.set(put_value, wait=True, timeout=timeout)
        reading = await signal.read()
        value = reading["test_signal"]["value"]
        if isinstance(value, np.ndarray):
            value = value.tolist()
        assert_close(value, put_value)

    dtype = py_type
    await _test_signal(dtype, proxy)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.skip("Not sure if we need tango_signal_auto")
@pytest.mark.parametrize(
    "pv, tango_type, d_format, py_type, initial_value, put_value, use_dtype, use_proxy",
    [
        (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
            use_dtype,
            use_proxy,
        )
        for (
            pv,
            tango_type,
            d_format,
            py_type,
            initial_value,
            put_value,
        ) in COMMANDS_SET
        for use_dtype in [True, False]
        for use_proxy in [True, False]
    ],
    ids=[
        f"{x[0]}_{use_dtype}_{use_proxy}"
        for x in COMMANDS_SET
        for use_dtype in [True, False]
        for use_proxy in [True, False]
    ],
)
async def test_tango_signal_auto_cmds(
    echo_device: str,
    pv: str,
    tango_type: str,
    d_format: AttrDataFormat,
    py_type: type[T],
    initial_value: T,
    put_value: T,
    use_dtype: bool,
    use_proxy: bool,
):
    source = echo_device + "/" + pv
    timeout = 0.2

    async def _test_signal(dtype, proxy):
        signal = await __tango_signal_auto(
            datatype=dtype,
            trl=source,
            device_proxy=proxy,
            name="test_signal",
            timeout=timeout,
        )
        # Ophyd SignalX does not support types
        assert type(signal) in [SignalR, SignalRW, SignalW]
        await signal.connect()
        assert signal
        reading = await signal.read()
        assert reading["test_signal"]["value"] is None

        await signal.set(put_value, wait=True, timeout=0.1)
        reading = await signal.read()
        value = reading["test_signal"]["value"]
        if isinstance(value, np.ndarray):
            value = value.tolist()
        assert_close(value, put_value)

    proxy = await DeviceProxy(echo_device) if use_proxy else None
    dtype = py_type if use_dtype else None
    await _test_signal(dtype, proxy)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.skip("Not sure if we need tango_signal_auto")
@pytest.mark.parametrize("use_proxy", [True, False])
async def test_tango_signal_auto_cmds_void(tango_test_device: str, use_proxy: bool):
    proxy = await DeviceProxy(tango_test_device) if use_proxy else None
    signal = await __tango_signal_auto(
        datatype=None,
        trl=tango_test_device + "/" + "clear",
        device_proxy=proxy,
    )
    assert type(signal) is SignalX
    await signal.connect()
    assert signal
    await signal.trigger(wait=True)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.skip("Not sure if we need tango_signal_auto")
async def test_tango_signal_auto_badtrl(tango_test_device: str):
    proxy = await DeviceProxy(tango_test_device)
    with pytest.raises(RuntimeError) as exc_info:
        await __tango_signal_auto(
            datatype=None,
            trl=tango_test_device + "/" + "badtrl",
            device_proxy=proxy,
        )
    assert f"Cannot find badtrl in {tango_test_device}" in str(exc_info.value)



import asyncio
from collections.abc import AsyncGenerator, AsyncIterator

from bluesky.protocols import StreamAsset
from event_model import DataKey

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorWriter,
    Device,
    DeviceVector,
    NameProvider,
    PathProvider,
    StrictEnum,
    observe_value,
    set_and_wait_for_value,
)
from ophyd_async.epics.core import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
)


class Writing(StrictEnum):
    ON = "ON"
    OFF = "OFF"


class OdinNode(Device):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.writing = epics_signal_r(Writing, f"{prefix}HDF:Writing")
        self.connected = epics_signal_r(bool, f"{prefix}Connected")

        super().__init__(name)


class Odin(Device):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.nodes = DeviceVector({i: OdinNode(f"{prefix}FP{i}:") for i in range(4)})

        self.capture = epics_signal_rw(
            Writing, f"{prefix}Writing", f"{prefix}ConfigHdfWrite"
        )
        self.num_captured = epics_signal_r(int, f"{prefix}FramesWritten")
        self.num_to_capture = epics_signal_rw_rbv(int, f"{prefix}ConfigHdfFrames")

        self.start_timeout = epics_signal_rw_rbv(int, f"{prefix}TimeoutTimerPeriod")

        self.image_height = epics_signal_rw_rbv(int, f"{prefix}DatasetDataDims0")
        self.image_width = epics_signal_rw_rbv(int, f"{prefix}DatasetDataDims1")

        self.num_row_chunks = epics_signal_rw_rbv(int, f"{prefix}DatasetDataChunks1")
        self.num_col_chunks = epics_signal_rw_rbv(int, f"{prefix}DatasetDataChunks2")

        self.file_path = epics_signal_rw_rbv(str, f"{prefix}ConfigHdfFilePath")
        self.file_name = epics_signal_rw_rbv(str, f"{prefix}ConfigHdfFilePrefix")

        self.acquisition_id = epics_signal_rw_rbv(
            str, f"{prefix}ConfigHdfAcquisitionId"
        )

        self.data_type = epics_signal_rw_rbv(str, f"{prefix}DatasetDataDatatype")

        super().__init__(name)


class OdinWriter(DetectorWriter):
    def __init__(
        self,
        path_provider: PathProvider,
        name_provider: NameProvider,
        odin_driver: Odin,
    ) -> None:
        self._drv = odin_driver
        self._path_provider = path_provider
        self._name_provider = name_provider
        super().__init__()

    async def open(self, multiplier: int = 1) -> dict[str, DataKey]:
        info = self._path_provider(device_name=self._name_provider())

        await asyncio.gather(
            self._drv.file_path.set(str(info.directory_path)),
            self._drv.file_name.set(info.filename),
            self._drv.data_type.set(
                "uint16"
            ),  # TODO: Get from eiger https://github.com/bluesky/ophyd-async/issues/529
            self._drv.num_to_capture.set(0),
        )

        await self._drv.capture.set(Writing.ON)

        return await self._describe()

    async def _describe(self) -> dict[str, DataKey]:
        data_shape = await asyncio.gather(
            self._drv.image_height.get_value(), self._drv.image_width.get_value()
        )

        return {
            "data": DataKey(
                source=self._drv.file_name.source,
                shape=list(data_shape),
                dtype="array",
                # TODO: Use correct type based on eiger https://github.com/bluesky/ophyd-async/issues/529
                dtype_numpy="<u2",
                external="STREAM:",
            )
        }

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        async for num_captured in observe_value(self._drv.num_captured, timeout):
            yield num_captured

    async def get_indices_written(self) -> int:
        return await self._drv.num_captured.get_value()

    def collect_stream_docs(self, indices_written: int) -> AsyncIterator[StreamAsset]:
        # TODO: Correctly return stream https://github.com/bluesky/ophyd-async/issues/530
        raise NotImplementedError()

    async def close(self) -> None:
        await set_and_wait_for_value(self._drv.capture, Writing.OFF)



import asyncio

import bluesky.plan_stubs as bps
import bluesky.plans as bp
from bluesky import RunEngine

from ophyd_async.tango.demo import (
    DemoCounter,
    DemoMover,
    TangoDetector,
)
from tango.test_context import MultiDeviceTestContext

content = (
    {
        "class": DemoMover,
        "devices": [{"name": "demo/motor/1"}],
    },
    {
        "class": DemoCounter,
        "devices": [{"name": "demo/counter/1"}, {"name": "demo/counter/2"}],
    },
)

tango_context = MultiDeviceTestContext(content)


async def main():
    with tango_context:
        detector = TangoDetector(
            trl="",
            name="detector",
            counters_kwargs={"prefix": "demo/counter/", "count": 2},
            mover_kwargs={"trl": "demo/motor/1"},
        )
        await detector.connect()

        RE = RunEngine()

        RE(bps.read(detector))
        RE(bps.mv(detector, 0))
        RE(bp.count(list(detector.counters.values())))

        set_status = detector.set(1.0)
        await asyncio.sleep(0.1)
        stop_status = detector.stop()
        await set_status
        await stop_status
        assert all([set_status.done, stop_status.done])
        assert all([set_status.success, stop_status.success])


if __name__ == "__main__":
    asyncio.run(main())



from collections import defaultdict

from bluesky import plans as bp
from bluesky.run_engine import RunEngine

from ophyd_async.core import assert_emitted
from ophyd_async.plan_stubs import ensure_connected
from ophyd_async.sim.demo import PatternDetector


# NOTE the async operations with h5py are non-trival
# because of lack of native support for async operations
# see https://github.com/h5py/h5py/issues/837
async def test_streaming_plan(RE: RunEngine, sim_pattern_detector: PatternDetector):
    names = []
    docs = []

    def append_and_print(name, doc):
        names.append(name)
        docs.append(doc)

    RE.subscribe(append_and_print)

    def plan():
        yield from ensure_connected(sim_pattern_detector, mock=True)
        yield from bp.count([sim_pattern_detector], num=1)

    RE(plan())

    # NOTE - double resource because double stream
    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_resource",
        "stream_datum",
        "stream_datum",
        "event",
        "stop",
    ]
    await sim_pattern_detector.writer.close()


async def test_plan(RE: RunEngine, sim_pattern_detector: PatternDetector):
    docs = defaultdict(list)

    def plan():
        yield from ensure_connected(sim_pattern_detector, mock=True)
        yield from bp.count([sim_pattern_detector])

    RE(plan(), lambda name, doc: docs[name].append(doc))

    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=2, event=1, stop=1
    )



from typing import Annotated as A

from ophyd_async.core import DEFAULT_TIMEOUT, AsyncStatus, SignalR, SignalRW, SignalX
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.tango import TangoReadable, tango_polling


# Enable device level polling, useful for servers that do not support events
# Polling for individual signal can be enabled with a dict
@tango_polling({"counts": (1.0, 0.1, 0.1), "sample_time": (0.1, 0.1, 0.1)})
class TangoCounter(TangoReadable):
    # Enter the name and type of the signals you want to use
    # If type is None or Signal, the type will be inferred from the Tango device
    counts: A[SignalR[int], Format.HINTED_SIGNAL]
    sample_time: A[SignalRW[float], Format.CONFIG_SIGNAL]
    start: SignalX
    reset_: SignalX

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        sample_time = await self.sample_time.get_value()
        timeout = sample_time + DEFAULT_TIMEOUT
        await self.start.trigger(wait=True, timeout=timeout)

    @AsyncStatus.wrap
    async def reset(self) -> None:
        await self.reset_.trigger(wait=True, timeout=DEFAULT_TIMEOUT)



from __future__ import annotations

import asyncio
import sys
from collections.abc import Coroutine, Iterator, Mapping, MutableMapping
from logging import LoggerAdapter, getLogger
from typing import Any, TypeVar
from unittest.mock import Mock

from bluesky.protocols import HasName
from bluesky.run_engine import call_in_bluesky_event_loop, in_bluesky_event_loop

from ._protocol import Connectable
from ._utils import DEFAULT_TIMEOUT, NotConnected, wait_for_connection

_device_mocks: dict[Device, Mock] = {}


class DeviceConnector:
    """Defines how a `Device` should be connected and type hints processed."""

    def create_children_from_annotations(self, device: Device):
        """Used when children can be created from introspecting the hardware.

        Some control systems allow introspection of a device to determine what
        children it has. To allow this to work nicely with typing we add these
        hints to the Device like so::

            my_signal: SignalRW[int]
            my_device: MyDevice

        This method will be run during ``Device.__init__``, and is responsible
        for turning all of those type hints into real Signal and Device instances.

        Subsequent runs of this function should do nothing, to allow it to be
        called early in Devices that need to pass references to their children
        during ``__init__``.
        """

    async def connect(
        self,
        device: Device,
        mock: bool | Mock,
        timeout: float,
        force_reconnect: bool,
    ):
        """Used during ``Device.connect``.

        This is called when a previous connect has not been done, or has been
        done in a different mock more. It should connect the Device and all its
        children.
        """
        coros = {}
        for name, child_device in device.children():
            child_mock = getattr(mock, name) if mock else mock  # Mock() or False
            coros[name] = child_device.connect(
                mock=child_mock, timeout=timeout, force_reconnect=force_reconnect
            )
        await wait_for_connection(**coros)


class Device(HasName, Connectable):
    """Common base class for all Ophyd Async Devices."""

    _name: str = ""
    #: The parent Device if it exists
    parent: Device | None = None
    # None if connect hasn't started, a Task if it has
    _connect_task: asyncio.Task | None = None
    # If not None, then this is the mock arg of the previous connect
    # to let us know if we can reuse an existing connection
    _connect_mock_arg: bool | None = None

    def __init__(
        self, name: str = "", connector: DeviceConnector | None = None
    ) -> None:
        self._connector = connector or DeviceConnector()
        self._connector.create_children_from_annotations(self)
        self.set_name(name)

    @property
    def name(self) -> str:
        """Return the name of the Device"""
        return self._name

    def children(self) -> Iterator[tuple[str, Device]]:
        for attr_name, attr in self.__dict__.items():
            if attr_name != "parent" and isinstance(attr, Device):
                yield attr_name, attr

    def set_name(self, name: str):
        """Set ``self.name=name`` and each ``self.child.name=name+"-child"``.

        Parameters
        ----------
        name:
            New name to set
        """
        self._name = name
        # Ensure self.log is recreated after a name change
        self.log = LoggerAdapter(
            getLogger("ophyd_async.devices"), {"ophyd_async_device_name": self.name}
        )
        for child_name, child in self.children():
            child_name = f"{self.name}-{child_name.strip('_')}" if self.name else ""
            child.set_name(child_name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "parent":
            if self.parent not in (value, None):
                raise TypeError(
                    f"Cannot set the parent of {self} to be {value}: "
                    f"it is already a child of {self.parent}"
                )
        elif isinstance(value, Device):
            value.parent = self
        return super().__setattr__(name, value)

    async def connect(
        self,
        mock: bool | Mock = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ) -> None:
        """Connect self and all child Devices.

        Contains a timeout that gets propagated to child.connect methods.

        Parameters
        ----------
        mock:
            If True then use ``MockSignalBackend`` for all Signals
        timeout:
            Time to wait before failing with a TimeoutError.
        """
        uses_mock = bool(mock)
        can_use_previous_connect = (
            uses_mock is self._connect_mock_arg
            and self._connect_task
            and not (self._connect_task.done() and self._connect_task.exception())
        )
        if mock is True:
            mock = Mock()  # create a new Mock if one not provided
        if force_reconnect or not can_use_previous_connect:
            self._connect_mock_arg = uses_mock
            if self._connect_mock_arg:
                _device_mocks[self] = mock
            coro = self._connector.connect(
                device=self, mock=mock, timeout=timeout, force_reconnect=force_reconnect
            )
            self._connect_task = asyncio.create_task(coro)

        assert self._connect_task, "Connect task not created, this shouldn't happen"
        # Wait for it to complete
        await self._connect_task


DeviceT = TypeVar("DeviceT", bound=Device)


class DeviceVector(MutableMapping[int, DeviceT], Device):
    """
    Defines device components with indices.

    In the below example, foos becomes a dictionary on the parent device
    at runtime, so parent.foos[2] returns a FooDevice. For example usage see
    :class:`~ophyd_async.epics.demo.DynamicSensorGroup`
    """

    def __init__(
        self,
        children: Mapping[int, DeviceT],
        name: str = "",
    ) -> None:
        self._children = dict(children)
        super().__init__(name=name)

    def __setattr__(self, name: str, child: Any) -> None:
        if name != "parent" and isinstance(child, Device):
            raise AttributeError(
                "DeviceVector can only have integer named children, "
                "set via device_vector[i] = child"
            )
        super().__setattr__(name, child)

    def __getitem__(self, key: int) -> DeviceT:
        return self._children[key]

    def __setitem__(self, key: int, value: DeviceT) -> None:
        # Check the types on entry to dict to make sure we can't accidentally
        # make a non-integer named child
        assert isinstance(key, int), f"Expected int, got {key}"
        assert isinstance(value, Device), f"Expected Device, got {value}"
        self._children[key] = value
        value.parent = self

    def __delitem__(self, key: int) -> None:
        del self._children[key]

    def __iter__(self) -> Iterator[int]:
        yield from self._children

    def __len__(self) -> int:
        return len(self._children)

    def children(self) -> Iterator[tuple[str, Device]]:
        for key, child in self._children.items():
            yield str(key), child

    def __hash__(self):  # to allow DeviceVector to be used as dict keys and in sets
        return hash(id(self))


class DeviceCollector:
    """Collector of top level Device instances to be used as a context manager

    Parameters
    ----------
    set_name:
        If True, call ``device.set_name(variable_name)`` on all collected
        Devices
    connect:
        If True, call ``device.connect(mock)`` in parallel on all
        collected Devices
    mock:
        If True, connect Signals in simulation mode
    timeout:
        How long to wait for connect before logging an exception

    Notes
    -----
    Example usage::

        [async] with DeviceCollector():
            t1x = motor.Motor("BLxxI-MO-TABLE-01:X")
            t1y = motor.Motor("pva://BLxxI-MO-TABLE-01:Y")
            # Names and connects devices here
        assert t1x.comm.velocity.source
        assert t1x.name == "t1x"

    """

    def __init__(
        self,
        set_name=True,
        connect=True,
        mock=False,
        timeout: float = 10.0,
    ):
        self._set_name = set_name
        self._connect = connect
        self._mock = mock
        self._timeout = timeout
        self._names_on_enter: set[str] = set()
        self._objects_on_exit: dict[str, Any] = {}

    def _caller_locals(self):
        """Walk up until we find a stack frame that doesn't have us as self"""
        try:
            raise ValueError
        except ValueError:
            _, _, tb = sys.exc_info()
            assert tb, "Can't get traceback, this shouldn't happen"
            caller_frame = tb.tb_frame
            while caller_frame.f_locals.get("self", None) is self:
                caller_frame = caller_frame.f_back
                assert (
                    caller_frame
                ), "No previous frame to the one with self in it, this shouldn't happen"
            return caller_frame.f_locals

    def __enter__(self) -> DeviceCollector:
        # Stash the names that were defined before we were called
        self._names_on_enter = set(self._caller_locals())
        return self

    async def __aenter__(self) -> DeviceCollector:
        return self.__enter__()

    async def _on_exit(self) -> None:
        # Name and kick off connect for devices
        connect_coroutines: dict[str, Coroutine] = {}
        for name, obj in self._objects_on_exit.items():
            if name not in self._names_on_enter and isinstance(obj, Device):
                if self._set_name and not obj.name:
                    obj.set_name(name)
                if self._connect:
                    connect_coroutines[name] = obj.connect(
                        self._mock, timeout=self._timeout
                    )

        # Connect to all the devices
        if connect_coroutines:
            await wait_for_connection(**connect_coroutines)

    async def __aexit__(self, type, value, traceback):
        self._objects_on_exit = self._caller_locals()
        await self._on_exit()

    def __exit__(self, type_, value, traceback):
        if in_bluesky_event_loop():
            raise RuntimeError(
                "Cannot use DeviceConnector inside a plan, instead use "
                "`yield from ophyd_async.plan_stubs.ensure_connected(device)`"
            )
        self._objects_on_exit = self._caller_locals()
        try:
            fut = call_in_bluesky_event_loop(self._on_exit())
        except RuntimeError as e:
            raise NotConnected(
                "Could not connect devices. Is the bluesky event loop running? See "
                "https://blueskyproject.io/ophyd-async/main/"
                "user/explanations/event-loop-choice.html for more info."
            ) from e
        return fut



from unittest.mock import Mock

import bluesky.plan_stubs as bps

from ophyd_async.core import DEFAULT_TIMEOUT, Device, wait_for_connection


def ensure_connected(
    *devices: Device,
    mock: bool | Mock = False,
    timeout: float = DEFAULT_TIMEOUT,
    force_reconnect=False,
):
    (connect_task,) = yield from bps.wait_for(
        [
            lambda: wait_for_connection(
                **{
                    device.name: device.connect(
                        mock=mock, timeout=timeout, force_reconnect=force_reconnect
                    )
                    for device in devices
                }
            )
        ]
    )

    if connect_task and connect_task.exception() is not None:
        raise connect_task.exception()



from ._detector import (
    DetectorController,
    DetectorTrigger,
    DetectorWriter,
    StandardDetector,
    TriggerInfo,
)
from ._device import Device, DeviceCollector, DeviceConnector, DeviceVector
from ._device_filler import DeviceFiller
from ._device_save_loader import (
    all_at_once,
    get_signal_values,
    load_device,
    load_from_yaml,
    save_device,
    save_to_yaml,
    set_signal_values,
    walk_rw_signals,
)
from ._flyer import FlyerController, StandardFlyer
from ._hdf_dataset import HDFDataset, HDFFile
from ._log import config_ophyd_async_logging
from ._mock_signal_backend import MockSignalBackend
from ._mock_signal_utils import (
    callback_on_mock_put,
    get_mock,
    get_mock_put,
    mock_puts_blocked,
    reset_mock_put_calls,
    set_mock_put_proceeds,
    set_mock_value,
    set_mock_values,
)
from ._protocol import AsyncConfigurable, AsyncReadable, AsyncStageable
from ._providers import (
    AutoIncrementFilenameProvider,
    AutoIncrementingPathProvider,
    DatasetDescriber,
    FilenameProvider,
    NameProvider,
    PathInfo,
    PathProvider,
    StaticFilenameProvider,
    StaticPathProvider,
    UUIDFilenameProvider,
    YMDPathProvider,
)
from ._readable import (
    ConfigSignal,
    HintedSignal,
    StandardReadable,
    StandardReadableFormat,
)
from ._signal import (
    Signal,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    assert_configuration,
    assert_emitted,
    assert_reading,
    assert_value,
    observe_value,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
)
from ._signal_backend import (
    Array1D,
    SignalBackend,
    SignalDatatype,
    SignalDatatypeT,
    make_datakey,
)
from ._soft_signal_backend import SignalMetadata, SoftSignalBackend
from ._status import AsyncStatus, WatchableAsyncStatus, completed_status
from ._table import Table
from ._utils import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    Callback,
    NotConnected,
    Reference,
    StrictEnum,
    SubsetEnum,
    T,
    WatcherUpdate,
    get_dtype,
    get_enum_cls,
    get_unique,
    in_micros,
    wait_for_connection,
)

__all__ = [
    "DetectorController",
    "DetectorTrigger",
    "DetectorWriter",
    "StandardDetector",
    "TriggerInfo",
    "Device",
    "DeviceConnector",
    "DeviceCollector",
    "DeviceVector",
    "DeviceFiller",
    "all_at_once",
    "get_signal_values",
    "load_device",
    "load_from_yaml",
    "save_device",
    "save_to_yaml",
    "set_signal_values",
    "walk_rw_signals",
    "StandardFlyer",
    "FlyerController",
    "HDFDataset",
    "HDFFile",
    "config_ophyd_async_logging",
    "MockSignalBackend",
    "callback_on_mock_put",
    "get_mock",
    "get_mock_put",
    "mock_puts_blocked",
    "reset_mock_put_calls",
    "set_mock_put_proceeds",
    "set_mock_value",
    "set_mock_values",
    "AsyncConfigurable",
    "AsyncReadable",
    "AsyncStageable",
    "AutoIncrementFilenameProvider",
    "AutoIncrementingPathProvider",
    "FilenameProvider",
    "NameProvider",
    "PathInfo",
    "PathProvider",
    "DatasetDescriber",
    "StaticFilenameProvider",
    "StaticPathProvider",
    "UUIDFilenameProvider",
    "YMDPathProvider",
    "ConfigSignal",
    "HintedSignal",
    "StandardReadable",
    "StandardReadableFormat",
    "Signal",
    "SignalR",
    "SignalRW",
    "SignalW",
    "SignalX",
    "assert_configuration",
    "assert_emitted",
    "assert_reading",
    "assert_value",
    "observe_value",
    "set_and_wait_for_value",
    "set_and_wait_for_other_value",
    "soft_signal_r_and_setter",
    "soft_signal_rw",
    "wait_for_value",
    "Array1D",
    "SignalBackend",
    "make_datakey",
    "StrictEnum",
    "SubsetEnum",
    "SignalDatatype",
    "SignalDatatypeT",
    "SignalMetadata",
    "SoftSignalBackend",
    "AsyncStatus",
    "WatchableAsyncStatus",
    "DEFAULT_TIMEOUT",
    "CalculatableTimeout",
    "Callback",
    "CALCULATE_TIMEOUT",
    "NotConnected",
    "Reference",
    "Table",
    "T",
    "WatcherUpdate",
    "get_dtype",
    "get_enum_cls",
    "get_unique",
    "in_micros",
    "wait_for_connection",
    "completed_status",
]



from ophyd_async.core import StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw_rbv


class PilatusTriggerMode(StrictEnum):
    internal = "Internal"
    ext_enable = "Ext. Enable"
    ext_trigger = "Ext. Trigger"
    mult_trigger = "Mult. Trigger"
    alignment = "Alignment"


class PilatusDriverIO(adcore.ADBaseIO):
    """This mirrors the interface provided by ADPilatus/db/pilatus.template."""

    def __init__(self, prefix: str, name: str = "") -> None:
        self.trigger_mode = epics_signal_rw_rbv(
            PilatusTriggerMode, prefix + "TriggerMode"
        )
        self.armed = epics_signal_r(bool, prefix + "Armed")
        super().__init__(prefix, name)



import os
import uuid
from datetime import date

import pytest

from ophyd_async.core import (
    AutoIncrementFilenameProvider,
    AutoIncrementingPathProvider,
    UUIDFilenameProvider,
    YMDPathProvider,
)


def test_auto_increment_filename_provider(static_path_provider_factory):
    auto_inc_fp = AutoIncrementFilenameProvider(inc_delimeter="")
    dp = static_path_provider_factory(auto_inc_fp)

    # Our filenames should go from 00000 to 99999.
    # We increment by one each time, so just check if casting filename to
    # int gets us i.
    for i in range(100000):
        info = dp()
        assert int(info.filename) == i

    # Since default max digits is 5, incrementing one more time to 100000
    # will go over the limit and raise a value error
    with pytest.raises(ValueError):
        dp()


@pytest.mark.parametrize(
    "uuid_version", [uuid.uuid1, uuid.uuid3, uuid.uuid4, uuid.uuid5]
)
def test_uuid_filename_provider(static_path_provider_factory, uuid_version):
    uuid_call_args = []
    if uuid_version in [uuid.uuid3, uuid.uuid5]:
        uuid_call_args = [uuid.NAMESPACE_URL, "test"]
    uuid_fp = UUIDFilenameProvider(
        uuid_call_func=uuid_version, uuid_call_args=uuid_call_args
    )

    dp = static_path_provider_factory(uuid_fp)

    info = dp()

    # Try creating a UUID object w/ given version to check if it is valid uuid
    uuid.UUID(info.filename, version=int(uuid_version.__name__[-1]))


@pytest.mark.parametrize("uuid_version", [uuid.uuid3, uuid.uuid5])
def test_uuid_filename_provider_no_namespace(
    static_path_provider_factory, uuid_version
):
    uuid_fp = UUIDFilenameProvider(uuid_call_func=uuid_version)
    dp = static_path_provider_factory(uuid_fp)

    with pytest.raises(ValueError):
        dp()


def test_auto_increment_path_provider(static_filename_provider, tmp_path):
    auto_inc_path_provider = AutoIncrementingPathProvider(
        static_filename_provider, tmp_path, num_calls_per_inc=3, increment=2
    )

    for _ in range(3):
        info = auto_inc_path_provider()
        assert os.path.basename(info.directory_path) == "00000"
    info = auto_inc_path_provider()
    assert os.path.basename(info.directory_path) == "00002"


def test_ymd_path_provider(static_filename_provider, tmp_path):
    ymd_path_provider = YMDPathProvider(static_filename_provider, tmp_path)
    current_date = date.today()
    date_path = (
        f"{current_date.year:04d}/{current_date.month:02d}/{current_date.day:02d}"
    )

    info_a = ymd_path_provider()
    assert info_a.directory_path == tmp_path / date_path

    info_b = ymd_path_provider(device_name="test_device")
    assert info_b.directory_path == tmp_path / "test_device" / date_path



from unittest.mock import patch

import pytest

from ophyd_async.core import DeviceCollector
from ophyd_async.sim.demo import PatternDetectorWriter, PatternGenerator


@pytest.fixture
async def writer(static_path_provider) -> PatternDetectorWriter:
    async with DeviceCollector(mock=True):
        driver = PatternGenerator()

    return PatternDetectorWriter(driver, static_path_provider, lambda: "NAME")


async def test_correct_descriptor_doc_after_open(writer: PatternDetectorWriter):
    with patch("ophyd_async.core._signal.wait_for_value", return_value=None):
        descriptor = await writer.open()

    assert descriptor == {
        "NAME": {
            "source": "sim://pattern-generator-hdf-file",
            "shape": [240, 320],
            "dtype": "array",
            "external": "STREAM:",
        },
        "NAME-sum": {
            "source": "sim://pattern-generator-hdf-file",
            "shape": [],
            "dtype": "number",
            "external": "STREAM:",
        },
    }

    await writer.close()


async def test_collect_stream_docs(writer: PatternDetectorWriter):
    await writer.open()
    [item async for item in writer.collect_stream_docs(1)]
    assert writer.pattern_generator._handle_for_h5_file



from pathlib import Path

from bluesky.utils import new_uid

from ophyd_async.core import (
    AsyncReadable,
    DeviceCollector,
    StandardFlyer,
    StaticFilenameProvider,
    StaticPathProvider,
)
from ophyd_async.epics import adsimdetector
from ophyd_async.sim.demo import SimMotor


async def make_detector(prefix: str, name: str, tmp_path: Path):
    fp = StaticFilenameProvider(f"test-{new_uid()}")
    dp = StaticPathProvider(fp, tmp_path)

    async with DeviceCollector(mock=True):
        det = adsimdetector.SimDetector(prefix, dp, name=name)

    det._config_sigs = [det.drv.acquire_time, det.drv.acquire]

    return det


async def test_readable():
    async with DeviceCollector(mock=True):
        det = await make_detector("test", "test det", Path("/tmp"))
    assert isinstance(SimMotor, AsyncReadable)
    assert isinstance(det, AsyncReadable)
    assert not isinstance(StandardFlyer, AsyncReadable)



import asyncio
import os
import time
from collections.abc import Callable, Sequence
from typing import Any

import numpy as np
import numpy.typing as npt
import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    SignalBackend,
    SoftSignalBackend,
    StrictEnum,
    T,
    soft_signal_rw,
)


class MyEnum(StrictEnum):
    a = "Aaa"
    b = "Bbb"
    c = "Ccc"


def integer_d(value):
    return {"dtype": "integer", "shape": []}


def number_d(value):
    return {"dtype": "number", "shape": []}


def string_d(value):
    return {"dtype": "string", "shape": []}


def enum_d(value):
    return {"dtype": "string", "shape": [], "choices": ["Aaa", "Bbb", "Ccc"]}


def waveform_d(value):
    return {"dtype": "array", "shape": [len(value)]}


class MonitorQueue:
    def __init__(self, backend: SignalBackend):
        self.backend = backend
        self.updates: asyncio.Queue[Reading] = asyncio.Queue()
        backend.set_callback(self.updates.put_nowait)

    async def assert_updates(self, expected_value):
        expected_reading = {
            "value": expected_value,
            "timestamp": pytest.approx(time.monotonic(), rel=0.1),
            "alarm_severity": 0,
        }
        reading = await self.updates.get()

        backend_value = await self.backend.get_value()
        backend_reading = await self.backend.get_reading()

        assert reading["value"] == expected_value == backend_value
        assert reading == expected_reading == backend_reading

    def close(self):
        self.backend.set_callback(None)


# Can be removed once numpy >=2 is pinned.
default_int_type = (
    "<i4" if os.name == "nt" and np.version.version.startswith("1.") else "<i8"
)


@pytest.mark.parametrize(
    "datatype, initial_value, put_value, descriptor, dtype_numpy",
    [
        (int, 0, 43, integer_d, default_int_type),
        (float, 0.0, 43.5, number_d, "<f8"),
        (str, "", "goodbye", string_d, "|S40"),
        (MyEnum, MyEnum.a, MyEnum.c, enum_d, "|S40"),
        (npt.NDArray[np.int8], [], [-8, 3, 44], waveform_d, "|i1"),
        (npt.NDArray[np.uint8], [], [218], waveform_d, "|u1"),
        (npt.NDArray[np.int16], [], [-855], waveform_d, "<i2"),
        (npt.NDArray[np.uint16], [], [5666], waveform_d, "<u2"),
        (npt.NDArray[np.int32], [], [-2], waveform_d, "<i4"),
        (npt.NDArray[np.uint32], [], [1022233], waveform_d, "<u4"),
        (npt.NDArray[np.int64], [], [-3], waveform_d, "<i8"),
        (npt.NDArray[np.uint64], [], [995444], waveform_d, "<u8"),
        (npt.NDArray[np.float32], [], [1.0], waveform_d, "<f4"),
        (npt.NDArray[np.float64], [], [0.2], waveform_d, "<f8"),
        (Sequence[str], [], ["nine", "ten"], waveform_d, "|S40"),
        # Can't do long strings until https://github.com/epics-base/pva2pva/issues/17
        # (str, "longstr", ls1, ls2, string_d),
        # (str, "longstr2.VAL$", ls1, ls2, string_d),
    ],
)
async def test_soft_signal_backend_get_put_monitor(
    datatype: type[T],
    initial_value: T,
    put_value: T,
    descriptor: Callable[[Any], dict],
    dtype_numpy: str,
):
    backend = SoftSignalBackend(datatype=datatype)

    await backend.connect(1)
    q = MonitorQueue(backend)
    try:
        # Check descriptor
        source = "soft://test"
        # Add expected dtype_numpy to descriptor
        assert dict(
            source=source, **descriptor(initial_value), dtype_numpy=dtype_numpy
        ) == await backend.get_datakey(source)
        # Check initial value
        await q.assert_updates(
            pytest.approx(initial_value) if initial_value != "" else initial_value
        )
        # Put to new value and check that
        await backend.put(put_value, True)
        await q.assert_updates(pytest.approx(put_value))
    finally:
        q.close()


async def test_soft_signal_backend_enum_value_equivalence():
    soft_backend = SoftSignalBackend(MyEnum)
    await soft_backend.connect(timeout=1)
    assert (await soft_backend.get_value()) is MyEnum.a
    await soft_backend.put(MyEnum.b, True)
    assert (await soft_backend.get_value()) is MyEnum.b


async def test_soft_signal_backend_with_numpy_typing():
    soft_backend = SoftSignalBackend(npt.NDArray[np.float64])
    await soft_backend.connect(timeout=1)
    array = await soft_backend.get_value()
    assert array.shape == (0,)


async def test_soft_signal_descriptor_fails_for_invalid_class():
    class myClass:
        def __init__(self) -> None:
            pass

    with pytest.raises(TypeError):
        SoftSignalBackend(myClass)


async def test_soft_signal_descriptor_with_metadata():
    soft_signal = soft_signal_rw(int, 0, units="mm", precision=0)
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert datakey[""]["units"] == "mm"
    assert datakey[""]["precision"] == 0

    soft_signal = soft_signal_rw(int, units="")
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert datakey[""]["units"] == ""
    assert not hasattr(datakey[""], "precision")


async def test_soft_signal_descriptor_with_no_metadata_not_passed():
    soft_signal = soft_signal_rw(int)
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert not hasattr(datakey[""], "units")
    assert not hasattr(datakey[""], "precision")


async def test_soft_signal_coerces_numpy_types():
    soft_signal = soft_signal_rw(float)
    await soft_signal.connect()
    assert await soft_signal.get_value() == 0.0
    assert type(await soft_signal.get_value()) is float
    await soft_signal.set(np.float64(1.1))
    assert await soft_signal.get_value() == 1.1
    assert type(await soft_signal.get_value()) is float
    soft_signal._connector.backend.set_value(np.float64(2.2))
    assert await soft_signal.get_value() == 2.2
    assert type(await soft_signal.get_value()) is float



import asyncio
from typing import Literal

from ophyd_async.core import (
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
    set_and_wait_for_value,
)
from ophyd_async.epics import adcore

from ._aravis_io import AravisDriverIO, AravisTriggerMode, AravisTriggerSource

# The deadtime of an ADaravis controller varies depending on the exact model of camera.
# Ideally we would maximize performance by dynamically retrieving the deadtime at
# runtime. See https://github.com/bluesky/ophyd-async/issues/308
_HIGHEST_POSSIBLE_DEADTIME = 1961e-6


class AravisController(DetectorController):
    GPIO_NUMBER = Literal[1, 2, 3, 4]

    def __init__(self, driver: AravisDriverIO, gpio_number: GPIO_NUMBER) -> None:
        self._drv = driver
        self.gpio_number = gpio_number
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return _HIGHEST_POSSIBLE_DEADTIME

    async def prepare(self, trigger_info: TriggerInfo):
        if trigger_info.total_number_of_triggers == 0:
            image_mode = adcore.ImageMode.continuous
        else:
            image_mode = adcore.ImageMode.multiple
        if (exposure := trigger_info.livetime) is not None:
            await self._drv.acquire_time.set(exposure)

        trigger_mode, trigger_source = self._get_trigger_info(trigger_info.trigger)
        # trigger mode must be set first and on it's own!
        await self._drv.trigger_mode.set(trigger_mode)

        await asyncio.gather(
            self._drv.trigger_source.set(trigger_source),
            self._drv.num_images.set(trigger_info.total_number_of_triggers),
            self._drv.image_mode.set(image_mode),
        )

    async def arm(self):
        self._arm_status = await set_and_wait_for_value(self._drv.acquire, True)

    async def wait_for_idle(self):
        if self._arm_status:
            await self._arm_status

    def _get_trigger_info(
        self, trigger: DetectorTrigger
    ) -> tuple[AravisTriggerMode, AravisTriggerSource]:
        supported_trigger_types = (
            DetectorTrigger.constant_gate,
            DetectorTrigger.edge_trigger,
            DetectorTrigger.internal,
        )
        if trigger not in supported_trigger_types:
            raise ValueError(
                f"{self.__class__.__name__} only supports the following trigger "
                f"types: {supported_trigger_types} but was asked to "
                f"use {trigger}"
            )
        if trigger == DetectorTrigger.internal:
            return AravisTriggerMode.off, AravisTriggerSource.freerun
        else:
            return (AravisTriggerMode.on, f"Line{self.gpio_number}")  # type: ignore

    async def disarm(self):
        await adcore.stop_busy_record(self._drv.acquire, False, timeout=1)



from __future__ import annotations

from unittest.mock import Mock

from ophyd_async.core import (
    Device,
    DeviceConnector,
    DeviceFiller,
    Signal,
    SignalR,
    SignalRW,
    SignalX,
)

from ._epics_connector import fill_backend_with_prefix
from ._signal import PvaSignalBackend, pvget_with_timeout

Entry = dict[str, str]


def _get_signal_details(entry: Entry) -> tuple[type[Signal], str, str]:
    match entry:
        case {"r": read_pv}:
            return SignalR, read_pv, read_pv
        case {"r": read_pv, "w": write_pv}:
            return SignalRW, read_pv, write_pv
        case {"rw": read_write_pv}:
            return SignalRW, read_write_pv, read_write_pv
        case {"x": execute_pv}:
            return SignalX, execute_pv, execute_pv
        case _:
            raise TypeError(f"Can't process entry {entry}")


class PviDeviceConnector(DeviceConnector):
    def __init__(self, prefix: str = "") -> None:
        # TODO: what happens if we get a leading "pva://" here?
        self.prefix = prefix
        self.pvi_pv = prefix + "PVI"

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            self.filler = DeviceFiller(
                device=device,
                signal_backend_factory=PvaSignalBackend,
                device_connector_factory=PviDeviceConnector,
            )
            # Devices will be created with unfilled PviDeviceConnectors
            list(self.filler.create_devices_from_annotations(filled=False))
            # Signals can be filled in with EpicsSignalSuffix and checked at runtime
            for backend, annotations in self.filler.create_signals_from_annotations(
                filled=False
            ):
                fill_backend_with_prefix(self.prefix, backend, annotations)
            self.filler.check_created()

    def _fill_child(self, name: str, entry: Entry, vector_index: int | None = None):
        if set(entry) == {"d"}:
            connector = self.filler.fill_child_device(name, vector_index=vector_index)
            connector.pvi_pv = entry["d"]
        else:
            signal_type, read_pv, write_pv = _get_signal_details(entry)
            backend = self.filler.fill_child_signal(name, signal_type, vector_index)
            backend.read_pv = read_pv
            backend.write_pv = write_pv

    async def connect(
        self, device: Device, mock: bool | Mock, timeout: float, force_reconnect: bool
    ) -> None:
        if mock:
            # Make 2 entries for each DeviceVector
            self.filler.create_device_vector_entries_to_mock(2)
        else:
            pvi_structure = await pvget_with_timeout(self.pvi_pv, timeout)
            entries: dict[str, Entry | list[Entry | None]] = pvi_structure[
                "value"
            ].todict()
            # Fill based on what PVI gives us
            for name, entry in entries.items():
                if isinstance(entry, dict):
                    # This is a child
                    self._fill_child(name, entry)
                else:
                    # This is a DeviceVector of children
                    for i, e in enumerate(entry):
                        if e:
                            self._fill_child(name, e, i)
            # Check that all the requested children have been filled
            self.filler.check_filled(f"{self.pvi_pv}: {entries}")
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect(device, mock, timeout, force_reconnect)
